#!/usr/bin/env python
#
# texture2d.py - The Texture2D class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Texture2D` and :class:`DepthTexture`
classes.
"""


import logging

import numpy     as np
import OpenGL.GL as gl

import fsl.transform.affine        as affine
import fsl.data.utils              as dutils
import fsleyes.gl                  as fslgl
import fsleyes.gl.shaders          as shaders
import fsleyes.gl.textures.texture as texture


log = logging.getLogger(__name__)


class DepthTexture(texture.Texture):
    """The ``DepthTexture`` class is a 2D ``GL_DEPTH_COMPONENT24`` texture
    which is used by the :class:`.RenderTexture` class.

    A ``DepthTexture`` is configured by setting its :meth:`.Texture.shape`
    property to the desired width/height.
    """


    def __init__(self, name):
        """Create a ``DepthTexture``

        :arg name: Unique name for this texture
        """
        texture.Texture.__init__(self, name, 2, 1, dtype=np.uint32)


    @property
    def dtype(self):
        """Overrides :meth:`.Texture.dtype`. """
        return np.uint32


    @dtype.setter
    def dtype(self):
        """Overrides the :meth:`.Texture.dtype` setter. Raises
        ``NotImplementedError``.
        """
        raise NotImplementedError()


    @property
    def textureType(self):
        """Overrides :meth:`.Texture.textureType`. """
        return gl.GL_UNSIGNED_INT


    @property
    def baseFormat(self):
        """Overrides :meth:`.Texture.baseFormat`. """
        return gl.GL_DEPTH_COMPONENT

    @property
    def internalFormat(self):
        """Overrides :meth:`.Texture.internalFormat`. """
        return gl.GL_DEPTH_COMPONENT24


    @texture.Texture.data.setter
    def data(self, data):
        """Overrides the :meth:`.Texture.data` setter. Raises an error -
        you cannot set data on a ``DepthTexture``.
        """
        raise NotImplementedError('Cannot set data on a DepthTexture')


    def doRefresh(self):
        """Refreshes this ``DepthTexture`` based on the current
        :meth:`.Texture.shape`.
        """

        width, height = self.shape
        ttype         = self.textureType
        intFmt        = self.internalFormat
        baseFmt       = self.baseFormat

        with self.bound():

            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT,   1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_COMPARE_MODE,
                               gl.GL_NONE)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MAG_FILTER,
                               gl.GL_NEAREST)
            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MIN_FILTER,
                               gl.GL_NEAREST)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_WRAP_S,
                               gl.GL_CLAMP_TO_EDGE)
            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_WRAP_T,
                               gl.GL_CLAMP_TO_EDGE)
            gl.glTexImage2D(gl.GL_TEXTURE_2D,
                            0,
                            intFmt,
                            width,
                            height,
                            0,
                            baseFmt,
                            ttype,
                            None)


    def getBitmap(self):
        """Returns the data stored in this ``DepthTexture`` as a ``numpy.uint8``
        array of shape ``(height, width, 3)``.

        The depth values are inverted and normalised, and duplicated across
        the three RGB channels.
        """

        intFmt        = self.baseFormat
        extFmt        = self.textureType
        ndtype        = self.dtype
        width, height = self.shape

        with self.bound():
            data = gl.glGetTexImage(gl.GL_TEXTURE_2D, 0, intFmt, extFmt, None)

        data = np.frombuffer(data, dtype=ndtype).astype(np.float32)
        data = (data.max() - data)
        data = (data - data.min()) / (data.max() - data.min())
        data = (data * 255).astype(np.uint8)

        data = data.reshape((height, width, 1))
        data = np.dstack((data, data, data))
        data = np.flipud(data)

        return data


class Texture2D(texture.Texture):
    """The ``Texture2D`` class represents a 2D texture. A ``Texture2D``
    instance can be used in one of two ways:

      - Setting the texture data via the :meth:`.Texture.data` method, and then
        drawing it to a scene via :meth:`draw` or :meth:`drawOnBounds`.

      - Setting the texture size via :meth:`.Texture.shape`, and then drawing
        to it by some other means (see e.g. the :class:`.RenderTexture` class,
        a sub-class of ``Texture2D``).
    """


    def __init__(self, name, **kwargs):
        """Create a ``Texture2D`` instance.

        :arg name: Unique name for this ``Texture2D``.
        """

        nvals = kwargs.pop('nvals', 4)

        if nvals not in (1, 3, 4):
            raise ValueError('nvals must be 1, 3 or 4')

        # We keep a copy of the current
        # width/height, so we can detect
        # whether it has changed, and
        # skip unnecessary processing
        self.__width  = None
        self.__height = None

        # If this texture is drawn via
        # the draw method, a shader
        # program is compiled and used.
        # The __shader method compiles
        # it and stores it here.
        self.__defaultShader = None

        texture.Texture.__init__(self, name, 2, nvals, **kwargs)


    def destroy(self):
        """Must be called when this ``Texture2D`` is no longer needed. """
        super().destroy()
        if self.__defaultShader is not None:
            self.__defaultShader.destroy()
            self.__defaultShader = None


    def doRefresh(self):
        """Overrides :meth:`.Texture.doRefresh`. Configures this ``Texture2D``.
        This includes setting up interpolation, and setting the texture size
        and data.
        """

        data = self.preparedData

        if   data is None:    width, height = self.shape
        elif self.nvals == 1: width, height = data.shape
        else:                 width, height = data.shape[1:]

        if data is not None:
            data = np.array(data.ravel('F'), copy=False)
            data = dutils.makeWriteable(data)

        interp = self.interp

        if interp is None:
            interp = gl.GL_NEAREST

        with self.bound():

            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT,   1)
            gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)

            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MAG_FILTER,
                               interp)
            gl.glTexParameteri(gl.GL_TEXTURE_2D,
                               gl.GL_TEXTURE_MIN_FILTER,
                               interp)

            if self.border is not None:
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_S,
                                   gl.GL_CLAMP_TO_BORDER)
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_T,
                                   gl.GL_CLAMP_TO_BORDER)
                gl.glTexParameterfv(gl.GL_TEXTURE_2D,
                                    gl.GL_TEXTURE_BORDER_COLOR,
                                    np.asarray(self.border, dtype=np.float32))
            else:
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_S,
                                   gl.GL_CLAMP_TO_EDGE)
                gl.glTexParameteri(gl.GL_TEXTURE_2D,
                                   gl.GL_TEXTURE_WRAP_T,
                                   gl.GL_CLAMP_TO_EDGE)

            # If the width and height have not
            # changed, then we don't need to
            # re-define the texture. But we can
            # use glTexSubImage2D if we have
            # data to upload
            if width  == self.__width  and \
               height == self.__height and \
               data is not None:
                gl.glTexSubImage2D(gl.GL_TEXTURE_2D,
                                   0,
                                   0,
                                   0,
                                   width,
                                   height,
                                   self.baseFormat,
                                   self.textureType,
                                   data)

            # If the width and/or height have
            # changed, we need to re-define
            # the texture properties
            else:
                self.__width  = width
                self.__height = height
                gl.glTexImage2D(gl.GL_TEXTURE_2D,
                                0,
                                self.internalFormat,
                                width,
                                height,
                                0,
                                self.baseFormat,
                                self.textureType,
                                data)


    def shapeData(self, data, oldShape=None):
        """Overrides :meth:`.Texture.shapeData`.

        This method is used by ``Texture2D`` sub-classes which are used to
        store 3D image data (e.g. the :class:`.ImageTexture2D` class). It
        shapes the data, ensuring that it is compatible with a 2D texture.

        :arg data:     ``numpy`` array containing the data
        :arg oldShape: Original data shape; if not provided, is taken from
                       ``data``.
        """

        nvals = self.nvals

        # For scalar, 1D or 2D data, we need
        # to make sure the data has a shape
        # compatible with the Texture2D
        if oldShape is None:
            oldShape = data.shape

        if nvals == 1:
            datShape = data.shape
        else:
            oldShape = oldShape[1:]
            datShape = data.shape[1:]

        oldShape = np.array(oldShape)
        datShape = np.array(datShape)

        if   np.all(oldShape         == [1, 1, 1]): newShape = ( 1,  1)
        elif np.all(oldShape[1:]     == [1, 1]):    newShape = (-1,  1)
        elif np.all(oldShape[[0, 2]] == [1, 1]):    newShape = (-1,  1)
        elif np.all(oldShape[:2]     == [1, 1]):    newShape = ( 1, -1)
        elif        oldShape[2]      == 1:          newShape = datShape[:2]
        elif        oldShape[1]      == 1:          newShape = datShape[[0, 2]]
        elif        oldShape[0]      == 1:          newShape = datShape[1:]

        if nvals > 1:
            newShape = [nvals] + list(newShape)

        return data.reshape(newShape)


    def texCoordXform(self, origShape):
        """Overrides :meth:`.Texture.texCoordXform`.

        Returns an affine matrix which encodes a rotation that maps the two
        major axes of the image voxel coordinate system to the first two axes
        of the texture coordinate system.

        This method is used by sub-classes which are being used to store 3D
        image data, e.g. the :class:`.ImageTexture2D` and
        :class:`.SelectionTexture2D` classes.

        If this texture does not have any data yet, this method will return
        ``None``.
        """

        scales  = [1, 1, 1]
        offsets = [0, 0, 0]
        rots    = [0, 0, 0]

        if origShape is None:
            return None

        # Here we apply a rotation to the
        # coordinates to force the two major
        # voxel axes to map to the first two
        # texture coordinate axes
        if origShape[0] == 1:
            rots      = [0, -np.pi / 2, -np.pi / 2]
        elif origShape[1] == 1:
            rots      = [-np.pi / 2, 0, 0]
            scales[1] = -1

        return affine.compose(scales, offsets, rots)


    def doPatch(self, data, offset):
        """Overrides :meth:`.Texture.doPatch`. Updates part of the texture
        data.
        """
        shape = data.shape
        data  = data.flatten(order='F')

        with self.bound():
            gl.glTexSubImage2D(gl.GL_TEXTURE_2D,
                               0,
                               offset[0],
                               offset[1],
                               shape[0],
                               shape[1],
                               self.baseFormat,
                               self.textureType,
                               data)


    def __shader(self):
        """Returns a shader program to draw the contents of this ``Texture2D``.
        Used by :meth:`draw`.
        """
        if self.__defaultShader is None:
            vertSrc   = shaders.getVertexShader(  'texture2d')
            fragSrc   = shaders.getFragmentShader('texture2d')

            if float(fslgl.GL_COMPATIBILITY) < 2.1:
                shader = shaders.ARBPShader(vertSrc, fragSrc, {'texture' : 0})
            else:
                shader = shaders.GLSLShader(vertSrc, fragSrc)
                with shader.loaded():
                    shader.set('tex', 0)
            self.__defaultShader = shader
        return self.__defaultShader


    def draw(self, vertices):
        """Draw the contents of this ``Texture2D`` to a region specified by
        the given vertices.
        """

        shader    = self.__shader()
        vertices  = np.array(vertices, dtype=np.float32)
        nverts    = vertices.shape[0]
        texCoords = self.generateTextureCoords()

        with self.bound(gl.GL_TEXTURE0), shader.loaded():
            shader.setAtt('vertex',   vertices)
            shader.setAtt('texCoord', texCoords)
            shader.draw(gl.GL_TRIANGLES, 0, nverts)


    def drawOnBounds(self,
                     zpos,
                     xmin,
                     xmax,
                     ymin,
                     ymax,
                     xax,
                     yax,
                     xform=None):
        """Draws the contents of this ``Texture2D`` to a rectangle.  This is a
        convenience method which creates a set of vertices, and passes them to
        the :meth:`draw` method.

        :arg zpos:  Position along the Z axis, in the display coordinate
                    system.
        :arg xmin:  Minimum X axis coordinate.
        :arg xmax:  Maximum X axis coordinate.
        :arg ymin:  Minimum Y axis coordinate.
        :arg ymax:  Maximum Y axis coordinate.
        :arg xax:   Display space axis which maps to the horizontal screen
                    axis.
        :arg yax:   Display space axis which maps to the vertical screen
                    axis.
        :arg xform: Transformation matrix to apply to the vertices.

        All other arguments are passed to the :meth:`draw` method.
        """

        vertices = self.generateVertices(
            zpos, xmin, xmax, ymin, ymax, xax, yax)
        if xform is not None:
            vertices = affine.transform(vertices, xform)
        self.draw(vertices)


    @classmethod
    def generateVertices(cls, zpos, xmin, xmax, ymin, ymax, xax, yax):
        """Generates a set of vertices suitable for passing to the
        :meth:`.Texture2D.draw` method, for drawing a ``Texture2D`` to a 2D
        canvas, with the ``GL_TRIANGLES`` primitive.

        :arg zpos:  Position along the Z axis, in the display coordinate
                    system.
        :arg xmin:  Minimum X axis coordinate.
        :arg xmax:  Maximum X axis coordinate.
        :arg ymin:  Minimum Y axis coordinate.
        :arg ymax:  Maximum Y axis coordinate.
        :arg xax:   Display space axis which maps to the horizontal screen
                    axis.
        :arg yax:   Display space axis which maps to the vertical screen
                    axis.
        """

        zax              = 3 - xax - yax
        vertices         = np.zeros((6, 3), dtype=np.float32)
        vertices[:, zax] = zpos

        vertices[ 0, [xax, yax]] = [xmin, ymin]
        vertices[ 1, [xax, yax]] = [xmin, ymax]
        vertices[ 2, [xax, yax]] = [xmax, ymin]
        vertices[ 3, [xax, yax]] = [xmax, ymin]
        vertices[ 4, [xax, yax]] = [xmin, ymax]
        vertices[ 5, [xax, yax]] = [xmax, ymax]

        return vertices


    @classmethod
    def generateTextureCoords(cls):
        """Generates a set of texture coordinates for drawing a
        :class:`Texture2D`. This function is used by the
        :meth:`Texture2D.draw` method.
        """

        texCoords       = np.zeros((6, 2), dtype=np.float32)
        texCoords[0, :] = [0, 0]
        texCoords[1, :] = [0, 1]
        texCoords[2, :] = [1, 0]
        texCoords[3, :] = [1, 0]
        texCoords[4, :] = [0, 1]
        texCoords[5, :] = [1, 1]

        return texCoords


    def getBitmap(self):
        """Returns the data stored in this ``Texture2D`` as a ``numpy.uint8``
        array of shape ``(height, width, 4)``.
        """

        intFmt        = self.baseFormat
        extFmt        = self.textureType
        ndtype        = self.dtype
        nvals         = self.nvals
        width, height = self.shape

        with self.bound():
            data = gl.glGetTexImage(gl.GL_TEXTURE_2D, 0, intFmt, extFmt, None)

        data = np.frombuffer(data, dtype=ndtype)
        data = data.reshape((height, width, nvals))
        data = np.flipud(data)

        return data



#!/usr/bin/env python
#
# editor.py - The Editor class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Editor` class, which provides
functionality to edit the data in an :class:`.Image` overlay.
"""


import logging

import collections.abc as abc

import numpy as np

import fsleyes.actions    as actions
from . import                selection


log = logging.getLogger(__name__)


class Editor(actions.ActionProvider):
    """The ``Editor`` class provides functionality to edit the data of an
    :class:`.Image` overlay. An ``Editor`` instance is associated with a
    specific ``Image`` overlay, passed to :meth:`__init__`.


    An ``Editor`` instance uses a :class:`.selection.Selection` object which
    allows voxel selections to be made, and keeps track of all changes to both
    the selection and image data.


    **The editing process**


    Making changes to the data in an :class:`.Image` involves two steps:

     1. Create a selection

     2. Change the value of voxels in that selection


    The first step can be peformed by working directly with the ``Selection``
    object - this is accessible via the :meth:`getSelection` method. The
    :meth:`fillSelection` method can be used to perform the second step.


    Some convenience methods are also provided for working with selections:

    .. autosummary::
       :nosignatures:

       getSelection
       fillSelection
       copySelection
       pasteSelection


    **Change tracking**


    An ``Editor`` instance keeps track of all changes made to the
    :class:`Image` data and to the ``Selection``  Every selection/data
    change made is recorded using :class:`SelectionChange` and
    :class:`.ValueChange` instances, which are stored in a list. These changes
    can be undone (and redone), through the :meth:`undo` and :meth:`redo`
    "action" methods (see the :mod:`.actions` module). Changes to the
    ``Selection`` object are, by default, only recorded when the selection is
    cleared. However, you can track *all* selection changes by initialising
    an ``Editor`` instance with ``recordSelection=True``.


    Sometimes it is useful to treat many small changes as a single large
    change.  For example, if a selection is being updated by dragging the
    mouse across a canvas, storing a separate change for every change in mouse
    position would result in many small changes which, if the user then wishes
    to undo, would have to be undone one by one. This problem can be overcome
    by the use of *change groups*. Whenever an operation similar to the above
    begins, you can call the :meth:`startChangeGroup` method - from now on,
    all changes will be aggregated into one group. When the operation
    completes, call the :meth:`endChangeGroup` to stop group changes.  When
    undoing/redoing changes, all of the changes in a change group will be
    undone/redone together.
    """


    def __init__(self,
                 image,
                 overlayList,
                 displayCtx,
                 recordSelection=False):
        """Create an ``Editor``.

        :arg image:           The :class:`.Image` instance being edited.

        :arg overlayList:     The :class:`.OverlayList` instance.

        :arg displayCtx:      The :class:`.DisplayContext` instance.

        :arg recordSelection: Defaults to ``False``. If ``True``, changes to
                              the :class:`.selection.Selection` are recorded
                              in the change history.
        """

        actions.ActionProvider.__init__(self, overlayList, displayCtx)

        self.__name      = '{}_{}'.format(self.__class__.__name__, id(self))
        self.__image     = image
        self.__selection = selection.Selection(
            image, displayCtx.getDisplay(image))

        if recordSelection:
            self.__selection.register(self.__name, self.__selectionChanged)

        # A list of state objects, providing
        # records of what has been done. The
        # doneIndex points to the current
        # state. Everything before the doneIndex
        # represents previous states, and
        # everything after the doneIndex
        # represents states which have been
        # undone.
        self.__doneList        = []
        self.__doneIndex       = -1
        self.__inGroup         = False
        self.__recordChanges   = True
        self.__recordSelection = recordSelection
        self.undo.enabled      = False
        self.redo.enabled      = False

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message."""
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def destroy(self):
        """Removes some property listeners, and clears references to objects
        to prevent memory leaks.
        """
        actions.ActionProvider.destroy(self)
        self.__selection.deregister(self.__name)
        self.__image     = None
        self.__selection = None
        self.__doneList  = None


    def getImage(self):
        """Returns the :class:`~fsl.data.image.Image` associated with this
        ``Editor``.
        """
        return self.__image


    def getSelection(self):
        """Returns the :class:`.selection.Selection` instance currently in use.
        """
        return self.__selection


    def clearSelection(self, *args, **kwargs):
        """Clears the :class:`.selection.Selection` (see
        :meth:`.selection.Selection.clearSelection`). If this ``Editor`` is
        not recording all selection changes (``recordSelection=False`` in
        :meth:`__init__`), the selection state before being cleared is saved
        in the change history.

        All arguments are passed through to :meth:`.Selection.clearSelection`.
        """
        self.__selection.clearSelection(*args, **kwargs)

        if (not self.__recordSelection) and self.__recordChanges:

            old, new, offset = self.__selection.getLastChange()

            # getLastChange will return None
            # if there is no last change
            if old is not None:
                change = SelectionChange(self.__image, offset, old, new)
                self.__changeMade(change)


    def invertSelection(self):
        """Inverts the current selection. """

        image    = self.__image
        sel      = self.__selection.getSelection()
        inverted = np.array(sel == 0, dtype=np.uint8)

        change = SelectionChange(image, (0, 0, 0), sel, inverted)

        self.__applyChange(change)
        self.__changeMade( change)


    def fillSelection(self, newVals):
        """Fills the current selection with the specified value or values.

        :arg newVals: A scalar value, or a sequence containing the same
                      number of values as the current selection size.
        """

        image               = self.__image
        opts                = self.displayCtx.getOpts(image)
        selectBlock, offset = self.__selection.getBoundedSelection()

        if not isinstance(newVals, abc.Sequence):
            nv = np.zeros(selectBlock.shape, dtype=np.float32)
            nv.fill(newVals)
            newVals = nv
        else:
            newVals = np.array(newVals)

        xlo, ylo, zlo = offset
        xhi = xlo + selectBlock.shape[0]
        yhi = ylo + selectBlock.shape[1]
        zhi = zlo + selectBlock.shape[2]

        slc     = opts.index((slice(xlo, xhi),
                              slice(ylo, yhi),
                              slice(zlo, zhi)))
        oldVals = image[slc]

        selectBlock = selectBlock == 0
        newVals[selectBlock] = oldVals[selectBlock]

        oldVals = np.array(oldVals)
        change  = ValueChange(image, opts.volume, offset, oldVals, newVals)

        self.__applyChange(change)
        self.__changeMade( change)


    def startChangeGroup(self):
        """Starts a change group. All subsequent changes will be grouped
        together, for :meth:`undo`/:meth:`redo` purposes, until a call to
        :meth:`endChangeGroup`.
        """

        if not self.__recordChanges:
            return

        del self.__doneList[self.__doneIndex + 1:]

        self.__inGroup    = True
        self.__doneIndex += 1
        self.__doneList.append([])

        log.debug('{}: starting change group - merging subsequent '
                  'changes at index {} of {}'.format(self.__image.name,
                                                     self.__doneIndex,
                                                     len(self.__doneList)))


    def endChangeGroup(self):
        """Ends a change group previously started by a call to
        :meth:`startChangeGroup`.
        """

        if not self.__recordChanges:
            return

        self.__inGroup = False
        log.debug('{}: ending change group at {} of {}'.format(
            self.__image.name,
            self.__doneIndex,
            len(self.__doneList)))


    def recordChanges(self, record=True):
        """Cause this ``Editor`` to either record or ignore any changes that
        are made to the selection or the image data until further notice.

        :arg record: If ``True``, changes are recorded. Otherwise they are
                     ignored.
        """

        self.__recordChanges = record

        if self.__recordSelection:
            self.__selection.enable(self.__name, enable=record)


    def ignoreChanges(self):
        """Cause this ``Editor`` to ignore any changes that are made to the
        selection or the image data until further notice. Call the
        :meth:`recordChanges` method to resume recording changes.
        """
        self.recordChanges(False)


    @actions.action
    def undo(self):
        """Un-does the most recent change.  Returns a list containing all
        change objects that were undone - either :class:`ValueChange` or
        :class:`SelectionChange` objects.
        """
        if self.__doneIndex == -1:
            return []

        log.debug('{}: undo change {} of {}'.format(
            self.__image.name,
            self.__doneIndex,
            len(self.__doneList)))

        change = self.__doneList[self.__doneIndex]

        if not isinstance(change, abc.Sequence):
            change = [change]

        for c in reversed(change):
            self.__revertChange(c)

        self.__doneIndex -= 1

        self.__inGroup = False
        self.redo.enabled = True
        if self.__doneIndex == -1:
            self.undo.enabled = False

        return change


    @actions.action
    def redo(self):
        """Re-does the most recent undone change.  Returns a list containing
        all change objects that were undone - either :class:`ValueChange` or
        :class:`SelectionChange` objects.
        """
        if self.__doneIndex == len(self.__doneList) - 1:
            return []

        log.debug('{}: redo change {} of {}'.format(
            self.__image.name,
            self.__doneIndex + 1,
            len(self.__doneList)))

        change = self.__doneList[self.__doneIndex + 1]

        if not isinstance(change, abc.Sequence):
            change = [change]

        for c in change:
            self.__applyChange(c)

        self.__doneIndex += 1

        self.__inGroup = False
        self.undo.enabled = True
        if self.__doneIndex == len(self.__doneList) - 1:
            self.redo.enabled = False

        return change


    def copySelection(self):
        """Copies the ``Image`` data in the current selection. Returns the
        data in a format that can be passed directly to the
        :meth:`pasteSelection` method of this, or another, ``Editor`` instance.

        .. note:: The format of the returned data might change, so I haven't
                  specified it.
        """

        image     = self.__image
        opts      = self.displayCtx.getOpts(image)
        selection = np.array(self.__selection.getSelection() > 0)
        data      = np.zeros(image.shape[:3], dtype=image.dtype)

        data[selection] = image[opts.index()][selection]

        return data, selection


    def pasteSelection(self, clipboard):
        """Pastes the data in the given ``clipboard`` into the ``Image`` that
        is managed by this ``Editor``.

        The ``clipboard`` is assumed to have been created by the
        :meth:`copySelection` method of another ``Editor`` instance which is
        managing an ``Image`` that has the same resolution and dimensions as
        the ``Image`` managed by this instance.
        """

        data, sel = clipboard
        self.__image[sel] = data[sel]


    def __selectionChanged(self, *a):
        """Called when the current :attr:`.Selection.selection` changes.

        Saves a record of the change with a :class:`SelectionChange` object.
        """

        if not self.__recordChanges:
            return

        old, new, offset = self.__selection.getLastChange()

        if old is not None:
            change = SelectionChange(self.__image, offset, old, new)
            self.__changeMade(change)


    def __changeMade(self, change):
        """Called by the :meth:`fillSelection` and :meth:`__selectionChanged`
        methods, whenever a data/selection change is made.

        Saves the change, and updates the state of the :meth:`undo`/
        :meth:`redo` methods.
        """

        if not self.__recordChanges:
            return

        if self.__inGroup:
            self.__doneList[self.__doneIndex].append(change)
        else:
            del self.__doneList[self.__doneIndex + 1:]
            self.__doneList.append(change)
            self.__doneIndex += 1

        self.undo.enabled = True
        self.redo.enabled = False

        log.debug('{}: new change to {} ({} of {})'.format(
            self.__image.name,
            change.overlay.name,
            self.__doneIndex,
            len(self.__doneList)))


    def __applyChange(self, change):
        """Called by the :meth:`fillSelection`  and :meth:`redo` methods.

        Applies the given ``change`` (either a :class:`ValueChange` or a
        :class:`SelectionChange`).
        """

        image = change.overlay
        opts  = self.displayCtx.getOpts(image)

        if isinstance(change, ValueChange):
            log.debug('%s: changing image %s data - offset '
                      '%s, volume %s, size %s',
                      self.__image.name,
                      change.overlay.name,
                      change.offset,
                      change.volume,
                      change.oldVals.shape)

            sliceobj = self.__makeSlice(change.offset,
                                        change.newVals.shape,
                                        opts.index()[3:])
            image[sliceobj] = change.newVals

        elif isinstance(change, SelectionChange):
            recording = self.__recordSelection
            if recording: self.__selection.disable(self.__name)
            self.__selection.setSelection(change.newSelection, change.offset)
            if recording: self.__selection.enable(self.__name)


    def __revertChange(self, change):
        """Called by the :meth:`undo` method. Reverses the change made by the
        given ``change`` object, (either a :class:`ValueChange` or a
        :class:`SelectionChange`)
        """

        image = change.overlay
        opts  = self.displayCtx.getOpts(image)

        if isinstance(change, ValueChange):
            log.debug('{}: reverting image {} data change - offset '
                      '{}, volume {}, size {}'.format(
                          self.__image.name,
                          change.overlay.name,
                          change.offset,
                          change.volume,
                          change.oldVals.shape))

            sliceobj = self.__makeSlice(change.offset,
                                        change.oldVals.shape,
                                        opts.index()[3:])
            image[sliceobj] = change.oldVals

        elif isinstance(change, SelectionChange):
            recording = self.__recordSelection
            if recording: self.__selection.disable(self.__name)
            self.__selection.setSelection(change.oldSelection, change.offset)
            if recording: self.__selection.enable(self.__name)


    def __makeSlice(self, offset, shape, volume=None):
        """Generate a tuple of ``slice`` objects and/or integers, suitable for
        indexing a region of an image at the given ``offset``, with the given
        ``shape``. If the image has more than three dimensions, the generated
        slice will index the specified ``volume`` (assumed to be a sequence of
        indices).
        """

        sliceobjs = []

        offset = [int(o) for o in offset]
        shape  = [int(s) for s in shape]

        for i in range(len(offset)):
            sliceobjs.append(slice(offset[i], offset[i] + shape[i], 1))

        if volume is not None:
            sliceobjs.extend(volume)

        return tuple(sliceobjs)


class ValueChange(object):
    """Represents a change which has been made to the data for an
    :class:`.Image` instance. Stores the location, the old values,
    and the new values.
    """


    def __init__(self, overlay, volume, offset, oldVals, newVals):
        """Create a ``ValueChange``.

        :arg overlay: The :class:`.Image` instance.
        :arg volume:  Sequence of volume indices, if ``overlay`` has more
                      than 3 dimensions.
        :arg offset:  Location (voxel coordinates) of the change.
        :arg oldVals: A ``numpy`` array containing the old image values.
        :arg newVals: A ``numpy`` array containing the new image values.
        """

        self.overlay = overlay
        self.volume  = volume
        self.offset  = offset
        self.oldVals = oldVals
        self.newVals = newVals


class SelectionChange(object):
    """Represents a change which has been made to a
    :class:`.selection.Selection` instance. Stores the location, the old
    selection, and the new selection.
    """


    def __init__(self, overlay, offset, oldSelection, newSelection):
        """Create a ``SelectionChange``.

        :arg overlay:      The :class:`.Image` instance.
        :arg offset:       Location (voxel coordinates) of the change.
        :arg oldSelection: A ``numpy`` array containing the old selection.
        :arg newSelection: A ``numpy`` array containing the new selection.
        """

        self.overlay      = overlay
        self.offset       = offset
        self.oldSelection = oldSelection
        self.newSelection = newSelection



#!/usr/bin/env python
#
# __init__.py - Funtions for managing OpenGL shader programs.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``shaders`` package contains classes and functions for finding,
parsing, compiling, and managing OpenGL shader programs. The following
types of shader program are supported:


 - GLSL 1.20 vertex and fragment shaders.
 - GLSL 3.30 vertex, geometry, and fragment shaders.
 - ``ARB_vertex_program`` and ``ARB_fragment_program`` shader programs.


The :mod:`.glsl` and :mod:`.arbp` packages respectively define the
:class:`.GLSLShader` and :class:`.ARBPShader` classes, which may be
used to manage shader programs of the corresponding type.


Some package-level functions are defined here, for finding and loading
shader source code:


  .. autosummary::
     :nosignatures:

     getShaderDir
     getShaderSuffix
     getVertexShader
     getGeometryShader
     getFragmentShader
"""


import os.path              as op

import                         fsleyes
import fsleyes.gl           as fslgl
from   .glsl import parse   as glslparse
from   .glsl import program as glslprogram
from   .arbp import program as arbpprogram


GLSLShader = glslprogram.GLSLShader
ARBPShader = arbpprogram.ARBPShader


def getShaderDir():
    """Returns the directory in which the ``ARB`` and ``glsl`` shader program
    source files can be found. A different directory will be returned depending
    on which OpenGL version is in use.
    """

    if   fslgl.GL_COMPATIBILITY == '3.3': subdir = 'gl33'
    if   fslgl.GL_COMPATIBILITY == '2.1': subdir = 'gl21'
    elif fslgl.GL_COMPATIBILITY == '1.4': subdir = 'gl14'

    return op.join(fsleyes.assetDir, 'gl', subdir)


def getFallbackShaderDir():
    """Return a fall-back shader source directory, used if a version-specific
    file does not exist. This is only used when ``GL_COMPATIBILITY`` is 3.3,
    as most of the GL 2.1 source files are re-used.
    """
    if fslgl.GL_COMPATIBILITY != '3.3':
        raise RuntimeError('No fallback shaders for GL version ' +
                           fslgl.GL_COMPATIBILITY)
    return op.join(fsleyes.assetDir, 'gl', 'gl21')


def getShaderSuffix():
    """Returns the shader program file suffix to use. A different suffix will be
    returned depending on which OpenGL version is in use.
    """

    if float(fslgl.GL_COMPATIBILITY) < 2.1: return 'prog'
    else:                                   return 'glsl'


def getVertexShader(prefix):
    """Returns the vertex shader source for the given GL type (e.g.
    'glvolume').
    """
    return _getShader(prefix, 'vert')


def getGeometryShader(prefix):
    """Returns the geometry shader source for the given GL type (e.g.
    'glvolume').
    """
    return _getShader(prefix, 'geom')


def getFragmentShader(prefix):
    """Returns the fragment shader source for the given GL type."""
    return _getShader(prefix, 'frag')


def _getShader(prefix, shaderType):
    """Returns the shader source for the given GL type and the given
    shader type (``'vert'``, ``'geom'``, or ``'frag'``).
    """
    shaderDir = getShaderDir()
    fname     = _getFileName(prefix, shaderType, shaderDir)

    # For gl33, we use shader files in the gl33 sub
    # dir if they exist, or fall back to gl21 if not,
    # running them through a converter to make them
    # gl33 compatbile.
    if not op.exists(fname):
        shaderDir   = getFallbackShaderDir()
        fname       = _getFileName(prefix, shaderType, shaderDir)

    with open(fname, 'rt', encoding='utf-8') as f:
        src = f.read()

    return preprocess(src, shaderType, shaderDir)


def _getFileName(prefix, shaderType, shaderDir):
    """Returns the file name of the shader program for the given GL type
    and shader type.
    """

    suffix = getShaderSuffix()

    if shaderType not in ('vert', 'geom', 'frag'):
        raise RuntimeError('Invalid shader type: {}'.format(shaderType))

    return op.join(shaderDir, '{}_{}.{}'.format(prefix, shaderType, suffix))


def preprocess(src, shaderType, shaderDir=None):
    """'Preprocess' the given shader source.

    This amounts to:

      - searching for lines containing '#pragma include filename',
        and replacing those lines with the contents of the specified
        files.

      - If ``fsleyes.gl.GL_COMPATIBILITY == '3.3'``, and the shader source
        is GLSL version 120 (from the ``'2.1'`` shader sources),
        updating it to be compatible with GLSL 330.
    """

    if shaderDir is None:
        shaderDir = getShaderDir()

    lines = src.split('\n')

    pragmas = []
    for linei, line in enumerate(lines):
        if line.startswith('#pragma'):
            pragmas.append(linei)

    includes = []
    for linei in pragmas:

        line = lines[linei].split()

        if len(line) != 3:       continue
        if line[1] != 'include': continue

        includes.append((linei, line[2]))

    for linei, fname in includes:
        fname = op.join(shaderDir, fname)
        with open(fname, 'rt', encoding='utf-8') as f:
            lines[linei] = f.read()

    src = '\n'.join(lines)

    if fslgl.GL_COMPATIBILITY == '3.3' and '#version 120' in src:
        src = glslparse.convert120to330(src, shaderType)

    return src



#!/usr/bin/env python


import sys
import logging
import pkgutil
import importlib


log = logging.getLogger(__name__)


def apply():

    mod     = sys.modules[__name__]
    path    = mod.__path__
    name    = mod.__name__
    submods = pkgutil.iter_modules(path, f'{name}.')

    for subpath, subname, _ in  submods:
        log.debug('Applying patch: %s', subname)
        importlib.import_module(subname)



#!/usr/bin/env python
#
# highdpi.py - Functions for enabling retina / high DPI rendering on OpenGL
# canvases.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions for enabling retina / high DPI rendering on
OpenGL canvases. It is used by the :class:`.WXGLCanvasTarget` class.

Two functions are defined:

  - :func:`needsToggling` returns ``True`` or ``False``, indicating whether
    a call to :func:`toggle` is necessary.

  - :func:`toggle` enables / disables high DPI rendering for a specific
    :class:`.WXGLCanvasTarget`.

On macOS with wxPython < 4.1.0, retina support for OpenGL canvases must be
programmatically enabled. In all other configurations, :func:`needsToggling`
will return False, and :func:`toggle` will do nothing.
"""


import              ctypes
import functools as ft
import              platform


@ft.lru_cache()
def _objc_runtime():
    """Create and return a wrapper around the objective-c runtime, with
    only the bits that we need.
    """

    objc = ctypes.cdll.LoadLibrary(ctypes.util.find_library('objc'))
    _    = ctypes.cdll.LoadLibrary(ctypes.util.find_library('Foundation'))

    # Get the class of an object
    objc.object_getClass.argtypes = [ctypes.c_void_p]
    objc.object_getClass.restype  =  ctypes.c_void_p

    # Get a method associated with an instance
    objc.class_getInstanceMethod.restype  =  ctypes.c_void_p
    objc.class_getInstanceMethod.argtypes = [ctypes.c_void_p, ctypes.c_void_p]

    # Register a method given its name
    objc.sel_registerName.restype  =  ctypes.c_void_p
    objc.sel_registerName.restype  =  ctypes.c_void_p
    objc.sel_registerName.argtypes = [ctypes.c_char_p]

    # Send a message (a.k.a. call a method) on an object
    objc.objc_msgSend.restype  =  ctypes.c_void_p
    objc.objc_msgSend.argtypes = [ctypes.c_void_p, ctypes.c_void_p]

    return objc


@ft.lru_cache()
def needsToggling():
    """Return ``True`` if high-DPI scaling can be toggled, ``False``
    otherwise.

    Under GTK, high DPI support is not possible with wxPython < 4.0.7, as
    ``wx.Window.GetContentScaleFactor`` always returns 1. Under GTK and
    from wxPython 4.0.7 onwards, GL canvases are scaled automatically. This
    function therefore returns ``False`` in both of these cases.

    Under macOS and wxpython < 4.1.0, high DPI must be explicitly requested
    for GL canvases via a Cocoa API call. This can be done via the
    :meth:`toggle` function.

    Under macOS and with wxpython >= 4.1.0, GL canvases are scaled
    automatically, so in this case this function will return ``False``.
    """

    import wx

    if platform.system() != 'Darwin':
        return False

    wxver = getattr(wx, '__version__', '1.0.0')
    wxver = [int(v) for v in wxver.split('.')[:3]]

    return wxver < [4, 1, 0]


def toggle(canvas, enable=True):
    """Enables/disables high-DPI/retina resolution rendering on the given
    ``wx.glcanvas.GLCanvas`` object.
    """
    if not needsToggling():
        return

    objc     = _objc_runtime()
    hd       = canvas.GetHandle()
    selector = objc.sel_registerName(b'setWantsBestResolutionOpenGLSurface:')
    objc.objc_msgSend(hd, selector, ctypes.c_bool(enable))



#!/usr/bin/env python
#
# glimageobject.py - The GLImageObject class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLImageObject` class, a sub-class of
:class:`.GLObject`, and the base class for all OpenGL objects which display
data from :class:`.Nifti` overlays.
"""


import numpy     as np
import OpenGL.GL as gl

import fsl.data.image                      as fslimage
import fsl.transform.affine                as affine
import fsl.utils.memoize                   as memoize

import fsleyes.displaycontext.volume3dopts as volume3dopts
import fsleyes.colourmaps                  as fslcmaps
import fsleyes.gl.globject                 as globject
import fsleyes.gl.textures                 as textures
import fsleyes.gl.resources                as glresources
import fsleyes.gl.routines                 as glroutines


class GLImageObject(globject.GLObject):
    """The ``GLImageObject`` class is the base class for all GL representations
    of :class:`.Nifti` instances. It contains some convenience methods for
    drawing volumetric image data.

    Some useful methods for 2D rendering::

      .. autsummary::
         :nosignatures:

         frontFace
         generateVertices2D
         generateVoxelCoordinates2D

    Some useful methods for 3D rendering::

      .. autsummary::
         :nosignatures:

         generateVertices3D
         generateVoxelCoordinates3D
         get3DClipPlane
         clipPlaneVertices
         drawClipPlanes
    """


    def __init__(self, overlay, overlayList, displayCtx, threedee):
        """Create a ``GLImageObject``.

        :arg image:       A :class:`.Nifti` object.

        :arg overlayList: The :class`.OverlayList`.

        :arg displayCtx:  The :class:`.DisplayContext` object managing the
                          scene.

        :arg threedee:    Set up for 2D or 3D rendering.
        """

        globject.GLObject.__init__(
            self, overlay, overlayList, displayCtx, threedee)

        self.__name = 'GLImageObject_{}'.format(self.name)

        name = self.__name
        opts = self.opts

        # In 3D mode, when Volume3DOpts.showClipPlanes
        # is on, we create a unique random colour for
        # each displayed clipping plane.
        self.__clipPlaneColours = {}

        # If this is being shown in 3D,
        # we add some listeners. Some
        # geometry methods are memoized,
        # and we need to invalidate the
        # memoize cache when certain
        # display properties change.
        if self.threedee and isinstance(opts, volume3dopts.Volume3DOpts):

            kwargs = {'callback' : self.__boundsChanged, 'immediate' : True}

            # Invalidate vertices on bounds change
            opts.addListener('bounds',       name, **kwargs)
            opts.addListener('transform',    name, **kwargs)
            opts.addListener('displayXform', name, **kwargs)

            # Invalidate 3Dclipping on any clip chagnes
            kwargs['callback'] = self.__clip3DChanged
            opts.addListener('clipPosition',    name, **kwargs)
            opts.addListener('clipAzimuth',     name, **kwargs)
            opts.addListener('clipInclination', name, **kwargs)


    @property
    def image(self):
        """The :class:`.Nifti` being rendered by this ``GLImageObject``. This
        is equivalent to :meth:`.GLObject.overlay`.
        """
        return self.overlay


    def destroy(self):
        """Must be called when this ``GLImageObject`` is no longer needed.
        Removes some property listeners.
        """

        name = self.__name
        opts = self.opts

        opts.removeListener('bounds',          name)
        opts.removeListener('transform',       name)
        opts.removeListener('displayXform',    name)

        if self.threedee and isinstance(opts, volume3dopts.Volume3DOpts):
            opts.removeListener('clipPosition',    name)
            opts.removeListener('clipAzimuth',     name)
            opts.removeListener('clipInclination', name)

        globject.GLObject.destroy(self)


    @property
    def destroyed(self):
        """Returns ``True`` if :meth:`destroy` has been called, ``False``
        otherwise.
        """
        return self.image is None


    def getDisplayBounds(self):
        """Returns the bounds of the :class:`.Image` (see the
        :meth:`.DisplayOpts.bounds` property).
        """
        return (self.opts.bounds.getLo(),
                self.opts.bounds.getHi())


    def getDataResolution(self, xax, yax):
        """Returns a suitable screen resolution for rendering this
        ``GLImageObject`` in 2D.
        """

        import nibabel as nib

        image = self.image
        opts  = self.opts

        # Figure out a good display resolution
        # along each voxel dimension
        shape = np.array(image.shape[:3])

        # Figure out an approximate
        # correspondence between the
        # voxel axes and the display
        # coordinate system axes.
        xform = opts.getTransform('id', 'display')
        axes  = nib.orientations.aff2axcodes(
            xform, ((0, 0), (1, 1), (2, 2)))

        # Re-order the voxel resolutions
        # in the display space
        res = [shape[axes[0]], shape[axes[1]], shape[axes[2]]]

        return res


    def frontFace(self, canvas):
        """Convenience method for 2D rendering.

        Image slices are generally drawn onto a 2D plane which is parallel to
        the viewing plane (see the :class:`.GLVolume` class). If the canvas
        that is drawing this ``GLImageObject`` has adjusted the projection
        matrix (e.g. via the :attr:`.SliceCanvas.invertX` or
        :attr:`.SliceCanvas.invertY` properties), the front or back face of
        this plane may be facing the viewing plane.

        So if face-culling is desired, this method returns the face that
        is facing away from the viewing plane, i.e. the face that can safely
        be culled.

        .. note:: This will raise an error if called on a ``GLImageObject``
                  which is being drawn by anything other than a
                  :class:`.SliceCanvas` or :class:`.LightBoxCanvas`.
        """

        front = gl.GL_CCW
        back  = gl.GL_CW

        numInverts = 0

        copts = canvas.opts

        if copts.invertX: numInverts += 1
        if copts.invertY: numInverts += 1

        if numInverts == 1:
            front, back = back, front

        return front


    def generateVertices2D(self, zpos, axes, bbox=None):
        """Generates vertex coordinates for a 2D slice of the :class:`.Image`,
        through the given ``zpos``, with the optional ``bbox`` applied to the
        coordinates.


        This is a convenience method for generating vertices which can be used
        to render a slice through a 3D texture. It is used by the
        :mod:`.gl14.glvolume_funcs` and :mod:`.gl21.glvolume_funcs` (and other)
        modules.


        A tuple of three values is returned, containing:

          - A ``6*3 numpy.float32`` array containing the vertex coordinates

          - A ``6*3 numpy.float32`` array containing the voxel coordinates
            corresponding to each vertex

          - A ``6*3 numpy.float32`` array containing the texture coordinates
            corresponding to each vertex
        """

        opts          = self.opts
        v2dMat        = opts.getTransform('voxel',   'display')
        d2vMat        = opts.getTransform('display', 'voxel')
        v2tMat        = opts.getTransform('voxel',   'texture')
        xax, yax, zax = axes

        vertices, voxCoords = glroutines.slice2D(
            self.image.shape[:3],
            xax,
            yax,
            zpos,
            v2dMat,
            d2vMat,
            bbox=bbox)

        # If not interpolating, centre the
        # voxel coordinates on the Z/depth
        # axis. We do this to avoid rounding
        # bias when the display Z position is
        # on a voxel boundary.
        if not hasattr(opts, 'interpolation') or opts.interpolation == 'none':
            voxCoords = opts.roundVoxels(voxCoords, daxes=[zax])

        texCoords = affine.transform(voxCoords, v2tMat)

        return vertices, voxCoords, texCoords


    @memoize.Instanceify(memoize.memoize)
    def generateVertices3D(self, bbox=None):
        """Generates vertex coordinates defining the 3D bounding box of the
        :class:`.Image`, with the optional ``bbox`` applied to the
        coordinates. See the :func:`.routines.boundingBox` function.

        A tuple of three values is returned, containing:

          - A ``36*3 numpy.float32`` array containing the vertex coordinates

          - A ``36*3 numpy.float32`` array containing the voxel coordinates
            corresponding to each vertex

          - A ``36*3 numpy.float32`` array containing the texture coordinates
            corresponding to each vertex
        """
        opts   = self.opts
        v2dMat = opts.getTransform('voxel',   'display')
        d2vMat = opts.getTransform('display', 'voxel')
        v2tMat = opts.getTransform('voxel',   'texture')

        vertices, voxCoords = glroutines.boundingBox(
            self.image.shape[:3],
            v2dMat,
            d2vMat,
            bbox=bbox)

        texCoords = affine.transform(voxCoords, v2tMat)

        return vertices, voxCoords, texCoords


    def generateVoxelCoordinates2D(
            self,
            zpos,
            axes,
            bbox=None,
            space='voxel'):
        """Generates a 2D grid of voxel coordinates along the
        XY display coordinate system plane, at the given ``zpos``.

        :arg zpos:  Position along the display coordinate system Z axis.

        :arg axes:  Axis indices.

        :arg bbox:  Limiting bounding box.

        :arg space: Either ``'voxel'`` (the default) or ``'display'``.
                    If the latter, the returned coordinates are in terms
                    of the display coordinate system. Otherwise, the
                    returned coordinates are integer voxel coordinates.

        :returns: A ``numpy.float32`` array of shape ``(N, 3)``, containing
                  the coordinates for ``N`` voxels.

        See the :func:`.pointGrid` function.
        """

        if space not in ('voxel', 'display'):
            raise ValueError('Unknown value for space ("{}")'.format(space))

        image         = self.image
        opts          = self.opts
        v2dMat        = opts.getTransform('voxel',   'display')
        d2vMat        = opts.getTransform('display', 'voxel')
        xax, yax, zax = axes

        # TODO If space == voxel, you should call
        #      pointGrid to generate voxels, and
        #      avoid the subsequent transform back
        #      from display to voxel space.

        if opts.transform == 'id':
            resolution = [1, 1, 1]
        else:
            resolution = image.pixdim[:3]

        voxels = glroutines.pointGrid(
            image.shape,
            resolution,
            v2dMat,
            xax,
            yax,
            bbox=bbox)[0]

        voxels[:, zax] = zpos

        if space == 'voxel':
            voxels = affine.transform(voxels, d2vMat)
            voxels = opts.roundVoxels(voxels,
                                      daxes=[zax],
                                      roundOther=False)

        return voxels


    def generateVoxelCoordinates3D(self, bbox, space='voxel'):
        """


        See the :func:`.pointGrid3D` function.

        note: Not implemented properly yet.
        """

        if space not in ('voxel', 'display'):
            raise ValueError('Unknown value for space ("{}")'.format(space))

        # TODO

        image      = self.image
        opts       = self.opts
        v2dMat     = opts.getTransform('voxel',   'display')
        d2vMat     = opts.getTransform('display', 'voxel')

        voxels = glroutines.pointGrid3D(image.shape[:3])

        if space == 'voxel':
            pass
            # voxels = affine.transform(voxels, d2vMat)
            # voxels = opts.roundVoxels(voxels)

        return voxels


    @memoize.Instanceify(memoize.memoize)
    def get3DClipPlane(self, *args, **kwargs):
        """Memoized wrapper around the :meth:`.Volume3DOpts.get3DClipPlane`
        method.
        """
        return self.opts.get3DClipPlane(*args, **kwargs)


    @memoize.Instanceify(memoize.memoize)
    def clipPlaneVertices(self, planeIdx, bbox=None):
        """A convenience method for use with overlays being displayed
        in terms of a :class:`.Volume3DOpts` instance.

        Generates vertices for the clipping plane specified by ``planeIdx``
        (an index into the ``Volume3DOpts.clip*`` lists).

        Returns a ``(N, 3)`` ``numpy`` array containing the vertices, and
        a 1D ``numpy`` array containing vertex indices.

        See the :meth:`get3DClipPlane` and :meth:`drawClipPlanes` methods.

        .. note:: This method depends on the ``trimesh`` library - if it is
                  not present, two empty arrays are returned.
        """

        origin, normal = self.get3DClipPlane(planeIdx)
        vertices       = self.generateVertices3D(bbox=bbox)[0]
        indices        = np.arange(vertices.shape[0]).reshape(-1, 3)

        try:
            import trimesh
            import trimesh.intersections as tmint
        except ImportError:
            return np.zeros((0, 3)), np.zeros((0,))

        # Calculate the intersection of the clip
        # plane with the image bounding box
        lines = tmint.mesh_plane(
            trimesh.Trimesh(vertices, indices),
            plane_normal=normal,
            plane_origin=origin)

        # I'm assuming that the
        # returned lines are sorted
        vertices = np.array(lines.reshape(-1, 3), dtype=np.float32)

        if vertices.shape[0] < 3:
            return np.zeros((0, 3)), np.zeros((0,))

        indices = glroutines.polygonIndices(vertices.shape[0])

        return vertices, indices


    def drawClipPlanes(self, xform=None, bbox=None):
        """A convenience method for use with overlays being displayed
        in terms of a :class:`.Volume3DOpts` instance.

        Draws the active clipping planes, as specified by the
        :class:`.Volume3DOpts` clipping properties.

        :arg xform: A transformation matrix to apply to the clip plane
                    vertices before drawing them.

        :arg bbox:  A bounding box by which the clip planes can be limited
                    (not currently honoured).
        """

        if not self.opts.showClipPlanes:
            return

        for i in range(self.opts.numClipPlanes):

            verts, idxs = self.clipPlaneVertices(i, bbox)

            if len(idxs) == 0:
                continue

            if xform is not None:
                verts = affine.transform(verts, xform)

            verts = np.array(verts.ravel('C'), dtype=np.float32, copy=False)

            # A consistent colour for
            # each clipping plane
            rgb = self.__clipPlaneColours.get(i, None)
            if rgb is None:
                rgb = fslcmaps.randomBrightColour()[:3]
                self.__clipPlaneColours[i] = rgb

            r, g, b = rgb

            with glroutines.enabled(gl.GL_VERTEX_ARRAY):

                gl.glColor4f(r, g, b, 0.3)
                gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
                gl.glDrawElements(gl.GL_TRIANGLES,
                                  len(idxs),
                                  gl.GL_UNSIGNED_INT,
                                  idxs)


    def __boundsChanged(self, *a):
        """Called when any change to the overlay bounds change.

        Some of the methods on this class use the
        :func:`fsl.utils.memoize.memoize` decorator to cache previously
        calculated values. When certain :class:`.DisplayOpts` properties
        change, these cached values need to be invalidated. This method
        does that.
        """

        self.generateVertices3D.invalidate()
        self.__clip3DChanged()


    def __clip3DChanged(self, *a):
        """Called when any change to the 3D clipping properties change.

        See the :meth:`__boundsChanged` method.
        """
        self.get3DClipPlane   .invalidate()
        self.clipPlaneVertices.invalidate()



#!/usr/bin/env python
#
# tractogramopts.py - The TractogramOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TractogramOpts` class, which defines
display properties for :class:`.Tractogram` overlays.
"""


import numpy   as np
import nibabel as nib

import fsl.data.image                       as fslimage
import fsl.transform.affine                 as affine
import fsleyes.gl                           as fslgl
import fsleyes.strings                      as strings
import fsleyes_props                        as props
import fsleyes.displaycontext.display       as fsldisplay
import fsleyes.displaycontext.colourmapopts as cmapopts
import fsleyes.displaycontext.vectoropts    as vectoropts


class TractogramOpts(fsldisplay.DisplayOpts,
                     cmapopts.ColourMapOpts,
                     vectoropts.VectorOpts):
    """Display options for :class:`.Tractogram` overlays. """


    colourMode = props.Choice(('orientation',))
    """Whether to colour streamlines by:
        - their orientation (e.g. RGB colouring)
        - per-vertex or per-streamline data
        - data from an image

    Per-vertex data sets and NIFTI images are dynamically added as options
    to this property.
    """


    clipMode = props.Choice((None,))
    """Whether to clip streamlines by:
        - per-vertex or per-streamline data
        - data from an image

    Per-vertex data sets and NIFTI images are dynamically added as options
    to this property.
    """


    lineWidth = props.Real(minval=1, maxval=10, default=2)
    """Width to draw the streamlines. When drawing in 3D, this controls the
    line width / tube diameter. When drawing in 2D, this controls the point
    diameter.
    """


    resolution = props.Int(minval=1, maxval=10, default=1, clamped=True)
    """When drawing in 3D as tubes, or in 2D as circles, this setting defines
    the resolution at which the tubes/circles are drawn. In 3D, if
    resolution <= 2, the streamlines are drawn as lines. In 2D, the
    resolution is clamped to a minimum of 3, with the effect that streamline
    vertices are drawn as triangles.
    """


    subsample = props.Percentage(default=100)
    """Draw a random sub-sample of all streamlines. This is useful when drawing
    very large tractograms.
    """


    def __init__(self, overlay, *args, **kwargs):
        """Create a ``TractogramOpts`` instance. """

        # Default to drawing a random sub-sample
        # of streamlines for large tractograms
        if overlay.nstreamlines > 150000:
            self.subsample = 15000000 / overlay.nstreamlines

        fsldisplay.DisplayOpts  .__init__(self, overlay, *args, **kwargs)
        cmapopts  .ColourMapOpts.__init__(self)
        vectoropts.VectorOpts   .__init__(self)

        olist         = self.overlayList
        lo, hi        = self.overlay.bounds
        xlo, ylo, zlo = lo
        xhi, yhi, zhi = hi
        self.bounds   = [xlo, xhi, ylo, yhi, zlo, zhi]

        self .addListener('colourMode', self.name, self.__colourModeChanged)
        self .addListener('clipMode',   self.name, self.__clipModeChanged)
        olist.addListener('overlays',   self.name, self.updateColourClipModes)

        self.updateColourClipModes()
        self.updateDataRange()


    def destroy(self):
        """Removes property listeners. """
        self.overlayList.removeListener('overlays', self.name)
        fsldisplay.DisplayOpts.destroy(self)


    @property
    def effectiveColourMode(self):
        """Returns one of ``'orientation'``, ``'vertexData'``, or
        ``'imageData'``, depending on the current :attr:`colourMode`.
        """
        ovl   = self.overlay
        cmode = self.colourMode
        if   isinstance(cmode, fslimage.Image): return 'imageData'
        elif cmode in ovl.vertexDataSets():     return 'vertexData'
        else:                                   return 'orientation'


    @property
    def effectiveClipMode(self):
        """Returns one of ``'none'``, ``'vertexData'``, or
        ``'imageData'``, depending on the current :attr:`clipMode`.
        """
        ovl   = self.overlay
        cmode = self.clipMode
        if   isinstance(cmode, fslimage.Image): return 'imageData'
        elif cmode in ovl.vertexDataSets():     return 'vertexData'
        else:                                   return 'none'


    def getLabels(self):
        """Overrides :meth:`.DisplayOpts.getLabele`. Returns orientation
        labels and codes for the coordinate system in which the streamline
        vertices are defined.
        """
        xorient, yorient, zorient  = self.overlay.orientation
        xlo = strings.anatomy['Nifti', 'lowshort',  xorient]
        ylo = strings.anatomy['Nifti', 'lowshort',  yorient]
        zlo = strings.anatomy['Nifti', 'lowshort',  zorient]
        xhi = strings.anatomy['Nifti', 'highshort', xorient]
        yhi = strings.anatomy['Nifti', 'highshort', yorient]
        zhi = strings.anatomy['Nifti', 'highshort', zorient]

        return ((xlo, ylo, zlo, xhi, yhi, zhi),
                (xorient, yorient, zorient))


    def getDataRange(self):
        """Overrides :meth:`.ColourMapOpts.getDataRange`. Returns the
        current data range to use for colouring - this depends on the
        current :attr:`colourMode`, and selected :attr:`vertexData` or
        :attr:`colourImage`.
        """
        data = self.__getData(self.colourMode)
        if data is None: return 0, 1
        else:            return np.nanmin(data), np.nanmax(data)


    def getClippingRange(self):
        """Overrides :meth:`.ColourMapOpts.getClippingRange`. Returns the
        current data range to use for clipping/thresholding - this depends on
        the selected :attr:`colourMode` and :attr:`vertexData` - if
        ``colourMode == 'orientation'``, the data may be clipped according
        to per-vertex data. Otherwise the clipping range will be equal to the
        display range.
        """

        colourMode = self.colourMode
        clipMode   = self.clipMode

        # if clipMode == colourMode, the same
        # data set that is used for colouring
        # will also be used for clipping
        if clipMode in (None, colourMode):
            return None

        data = self.__getData(clipMode)

        if data is None: return None
        else:            return np.nanmin(data), np.nanmax(data)


    def updateColourClipModes(self, *_):
        """Called when the :class:`.OverlayList` changes, and may be called
        externally (see e.g. :func:`.loadvertexdata.loadVertexData`) .
        Refreshes the options available on the :attr:`colourMode` and
        :attr:`clipMode` properties - ``'orientation'``, all vertex data
        sets on the :class:`.Tractogram` overlay, and all :class:`.Image`
        overlays in the :class:`.OverlayList`.
        """

        overlay    = self.overlay
        colourProp = self.getProp('colourMode')
        colour     = self.colourMode
        clipProp   = self.getProp('clipMode')
        clip       = self.clipMode

        vdata     = overlay.vertexDataSets()
        overlays  = self.displayCtx.getOrderedOverlays()
        overlays  = [o for o in overlays if isinstance(o, fslimage.Image)]

        colourOptions = ['orientation'] + overlays + vdata
        clipOptions   = [None]          + overlays + vdata

        colourProp.setChoices(colourOptions, instance=self)
        clipProp  .setChoices(clipOptions,   instance=self)

        # Preserve previous value,
        # or revert to default
        if colour in colourOptions: self.colourMode = colour
        else:                       self.colourMode = 'orientation'
        if clip   in clipOptions:   self.clipMode   = clip
        else:                       self.clipMode   = None


    @property
    def displayTransform(self):
        """Return an affine transformation which will transform streamline
        vertex coordinates into the current display coordinate system.
        """
        ref = self.displayCtx.displaySpace

        if not isinstance(ref, fslimage.Image):
            return np.eye(4)

        opts = self.displayCtx.getOpts(ref)

        return opts.getTransform('world', 'display')


    def sliceWidth(self, zax):
        """Returns a width along the specified **display** coordinate system
        axis, to be used for drawing a 2D slice through the tractogram on the
        axis plane.
        """

        # The z axis is specified in terms of
        # the display coordinate system -
        # identify the corresponding axis in the
        # tractogram/world coordinate system.
        codes = [[0, 0], [1, 1], [2, 2]]
        xform = affine.invert(self.displayTransform)
        zax   = nib.orientations.aff2axcodes(xform, codes)[zax]

        los, his = self.overlay.bounds
        zlen     = his[zax] - los[zax]
        return zlen / 200


    def __colourModeChanged(self, *_):
        """Called when :attr:`colourMode` changes.  Calls
        :meth:`.ColourMapOpts.updateDataRange`, to ensure that the display
        and clipping ranges are up to date.
        """
        self.updateDataRange(resetCR=(self.clipMode is None))


    def __clipModeChanged(self, *_):
        """Called when :attr:`clipMode` changes.  Calls
        :meth:`.ColourMapOpts.updateDataRange`, to ensure that the display
        and clipping ranges are up to date.
        """
        self.updateDataRange(resetDR=False)


    def __getData(self, mode):
        """Used by :meth:`getDataRange` and :meth:`getClippingRange`. Returns
        a numpy array containing data to be used for colouring/clipping.

        :arg mode: Current value of :attr:`colourMode` or :attr:`clipMode`.
        """
        overlay = self.overlay

        if isinstance(mode, fslimage.Image):
            return mode.data
        elif mode in overlay.vertexDataSets():
            return overlay.getVertexData(mode)
        # mode == 'orientation', or an invalid value
        else:
            return None



#!/usr/bin/env python
#
# slicecanvas.py - The SliceCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SliceCanvas` class, which contains the
functionality to display a 2D slice from a collection of 3D overlays.
"""


import logging

import OpenGL.GL as gl

import numpy as np

import fsl.data.image                     as fslimage
import fsl.utils.idle                     as idle
import fsl.transform.affine               as affine
import fsleyes_widgets.utils.status       as status
import fsleyes_props                      as props

import fsleyes.strings                    as strings
import fsleyes.displaycontext.canvasopts  as canvasopts
import fsleyes.gl.routines                as glroutines
import fsleyes.gl.resources               as glresources
import fsleyes.gl.globject                as globject
import fsleyes.gl.annotations             as annotations


log = logging.getLogger(__name__)


class SliceCanvas:
    """The ``SliceCanvas`` represents a canvas which may be used to display a
    single 2D slice from a collection of 3D overlays.  See also the
    :class:`.LightBoxCanvas`, a sub-class of ``SliceCanvas``.


    .. note:: The :class:`SliceCanvas` class is not intended to be instantiated
              directly - use one of these subclasses, depending on your
              use-case:

               - :class:`.OSMesaSliceCanvas` for static off-screen rendering of
                 a scene using OSMesa.

               - :class:`.WXGLSliceCanvas` for interactive rendering on a
                 :class:`wx.glcanvas.GLCanvas` canvas.


    The ``SliceCanvas`` creates a :class:`.SliceCanvasOpts` instance to manage
    its settings. The scene scene displayed on a ``SliceCanvas`` instance can
    be manipulated via the properties of its ``SliceCanvasOpts`` instnace,
    which is accessed via the ``opts`` attribute.


    **GL objects**


    The ``SliceCanvas`` draws :class:`.GLObject` instances. When created, a
    ``SliceCanvas`` creates a :class:`.GLObject` instance for every overlay in
    the :class:`.OverlayList`. When an overlay is added or removed, it
    creates/destroys ``GLObject`` instances accordingly.  Furthermore,
    whenever the :attr:`.Display.overlayType` for an existing overlay
    changes, the ``SliceCanvas`` destroys the old ``GLObject`` associated with
    the overlay, and creates a new one.


    ``GLObject`` instances may be shared amongst multiple ``SliceCanvas``
    instances - this is achieved with the :mod:`.resources` module.


    The ``SliceCanvas`` also uses an :class:`.Annotations` instance, for
    drawing simple annotations on top of the overlays.  This ``Annotations``
    instance can be accessed with the :meth:`getAnnotations` method.


    **Attributes and methods**


    The following attributes are available on a ``SliceCanvas``:


    =============== ===========================================
    ``name``        A unique name for this ``SliceCanvas``
    ``opts``        Reference to the :class:`.SliceCanvasOpts`.
    ``overlayList`` Reference to the :class:`.OverlayList`.
    ``displayCtx``  Reference to the :class:`.DisplayContext`.
    =============== ===========================================


    The following convenience methods are available on a ``SliceCanvas``:

    .. autosummary::
       :nosignatures:

       canvasToWorld
       worldToCanvas
       pixelSize
       panDisplayBy
       centreDisplayAt
       panDisplayToShow
       zoomTo
       resetDisplay
       getAnnotations
       viewport
       projectionMatrix
       viewMatrix
    """


    def __init__(self, overlayList, displayCtx, zax=0, opts=None):
        """Create a ``SliceCanvas``.

        :arg overlayList: An :class:`.OverlayList` object containing a
                          collection of overlays to be displayed.

        :arg displayCtx:  A :class:`.DisplayContext` object which describes
                          how the overlays should be displayed.

        :arg zax:         Display coordinate system axis perpendicular to the
                          plane to be displayed (the *depth* axis), default 0.
        """

        if opts is None:
            opts = canvasopts.SliceCanvasOpts()

        self.opts        = opts
        self.overlayList = overlayList
        self.displayCtx  = displayCtx
        self.name        = '{}_{}'.format(self.__class__.__name__, id(self))

        # A GLObject instance is created for
        # every overlay in the overlay list,
        # and stored in this dictionary
        self._glObjects = {}

        # Copies of the final viewport, modelView and
        # projection matrices are stored here on each
        # call to _draw.  They can be accessed via the
        # viewport, projectionMatrix, and viewMatrix
        # methods.
        self.__viewport         = None
        self.__projectionMatrix = None
        self.__viewMatrix       = None

        # The zax property is the image axis which
        # maps to the 'depth' axis of this canvas.
        opts.zax = zax

        # when any of the properties of this
        # canvas change, we need to redraw
        opts.addListener('zax',           self.name, self._zAxisChanged)
        opts.addListener('pos',           self.name, self.Refresh)
        opts.addListener('displayBounds', self.name, self.Refresh)
        opts.addListener('bgColour',      self.name, self.Refresh)
        opts.addListener('cursorColour',  self.name, self.Refresh)
        opts.addListener('cursorWidth',   self.name, self.Refresh)
        opts.addListener('showCursor',    self.name, self.Refresh)
        opts.addListener('cursorGap',     self.name, self.Refresh)
        opts.addListener('invertX',       self.name, self.Refresh)
        opts.addListener('invertY',       self.name, self.Refresh)
        opts.addListener('zoom',          self.name, self._zoomChanged)
        opts.addListener('highDpi',       self.name, self._highDpiChange)

        # When the overlay list changes, refresh the
        # display, and update the display bounds
        self.overlayList.addListener('overlays',
                                     self.name,
                                     self._overlayListChanged)
        self.displayCtx .addListener('overlayOrder',
                                     self.name,
                                     self.Refresh)
        self.displayCtx .addListener('bounds',
                                     self.name,
                                     self._overlayBoundsChanged)
        self.displayCtx .addListener('displaySpace',
                                     self.name,
                                     self._displaySpaceChanged)

        # The zAxisChanged method
        # will kick everything off
        self._annotations = annotations.Annotations(self)
        self._zAxisChanged()


    def destroy(self):
        """This method must be called when this ``SliceCanvas`` is no longer
        being used.

        It removes listeners from all :class:`.OverlayList`,
        :class:`.DisplayContext`, and :class:`.Display` instances, and
        destroys OpenGL representations of all overlays.
        """
        opts = self.opts
        opts.removeListener('zax',             self.name)
        opts.removeListener('pos',             self.name)
        opts.removeListener('displayBounds',   self.name)
        opts.removeListener('showCursor',      self.name)
        opts.removeListener('invertX',         self.name)
        opts.removeListener('invertY',         self.name)
        opts.removeListener('zoom',            self.name)
        opts.removeListener('highDpi',         self.name)

        self.overlayList.removeListener('overlays',     self.name)
        self.displayCtx .removeListener('bounds',       self.name)
        self.displayCtx .removeListener('displaySpace', self.name)
        self.displayCtx .removeListener('overlayOrder', self.name)

        for overlay in self.overlayList:
            disp  = self.displayCtx.getDisplay(overlay)
            globj = self._glObjects.get(overlay)

            disp.removeListener('overlayType',  self.name)
            disp.removeListener('enabled',      self.name)

            # globj could be None, or could
            # be False - see genGLObject.
            if globj:
                globj.deregister(self.name)
                glresources.delete(self.globjectId(overlay))

        self._annotations.destroy()

        self._annotations = None
        self.opts         = None
        self.overlayList  = None
        self.displayCtx   = None
        self._glObjects   = None


    @property
    def destroyed(self):
        """Returns ``True`` if a call to :meth:`destroy` has been made,
        ``False`` otherwise.
        """
        return self.overlayList is None


    def canvasToWorld(self, xpos, ypos, invertX=None, invertY=None):
        """Given pixel x/y coordinates on this canvas, translates them
        into xyz display coordinates.

        :arg invertX: If ``None``, taken from :attr:`.invertX`.
        :arg invertY: If ``None``, taken from :attr:`.invertY`.
        """

        opts = self.opts

        if invertX is None: invertX = opts.invertX
        if invertY is None: invertY = opts.invertY

        realWidth                 = opts.displayBounds.xlen
        realHeight                = opts.displayBounds.ylen
        canvasWidth, canvasHeight = [float(s) for s in self.GetSize()]

        if invertX: xpos = canvasWidth  - xpos
        if invertY: ypos = canvasHeight - ypos

        if realWidth    == 0 or \
           canvasWidth  == 0 or \
           realHeight   == 0 or \
           canvasHeight == 0:
            return None

        xpos = opts.displayBounds.xlo + (xpos / canvasWidth)  * realWidth
        ypos = opts.displayBounds.ylo + (ypos / canvasHeight) * realHeight

        pos = [None] * 3
        pos[opts.xax] = xpos
        pos[opts.yax] = ypos
        pos[opts.zax] = opts.pos[opts.zax]

        return pos


    def worldToCanvas(self, pos):
        """Converts a location in the display coordinate system into
        an x/y location in pixels relative to this ``SliceCanvas``.
        """

        opts    = self.opts
        xpos    = pos[opts.xax]
        ypos    = pos[opts.yax]
        invertX = opts.invertX
        invertY = opts.invertY

        xmin          = opts.displayBounds.xlo
        xlen          = opts.displayBounds.xlen
        ymin          = opts.displayBounds.ylo
        ylen          = opts.displayBounds.ylen
        width, height = [float(s) for s in self.GetSize()]

        if xlen   == 0 or \
           ylen   == 0 or \
           width  == 0 or \
           height == 0:
            return None

        xpos = width  * ((xpos - xmin) / xlen)
        ypos = height * ((ypos - ymin) / ylen)

        if invertX: xpos = width  - xpos
        if invertY: ypos = height - ypos

        return xpos, ypos


    def pixelSize(self):
        """Returns the current (x, y) size of one logical pixel in display
        coordinates.
        """
        w, h = self.GetSize()
        xlen = self.opts.displayBounds.xlen
        ylen = self.opts.displayBounds.ylen
        return (xlen / w, ylen / h)


    def panDisplayBy(self, xoff, yoff):
        """Pans the canvas display by the given x/y offsets (specified in
        display coordinates).
        """

        if len(self.overlayList) == 0: return

        xmin, xmax, ymin, ymax = self.opts.displayBounds[:]

        xmin = xmin + xoff
        xmax = xmax + xoff
        ymin = ymin + yoff
        ymax = ymax + yoff

        self.opts.displayBounds[:] = [xmin, xmax, ymin, ymax]


    def centreDisplayAt(self, xpos, ypos):
        """Pans the display so the given x/y position is in the centre. """

        xcentre, ycentre = self.getDisplayCentre()
        self.panDisplayBy(xpos - xcentre, ypos - ycentre)


    def getDisplayCentre(self):
        """Returns the horizontal/vertical position, in display coordinates,
        of the current centre of the display bounds.
        """
        bounds  = self.opts.displayBounds
        xcentre = bounds.xlo + (bounds.xhi - bounds.xlo) * 0.5
        ycentre = bounds.ylo + (bounds.yhi - bounds.ylo) * 0.5

        return xcentre, ycentre


    def panDisplayToShow(self, xpos, ypos):
        """Pans the display so that the given x/y position (in display
        coordinates) is visible.
        """

        bounds = self.opts.displayBounds

        # Do nothing if the position
        # is already being displayed
        if xpos >= bounds.xlo and xpos <= bounds.xhi and \
           ypos >= bounds.ylo and ypos <= bounds.yhi: return

        xoff = 0
        yoff = 0

        if   xpos <= bounds.xlo: xoff = xpos - bounds.xlo
        elif xpos >= bounds.xhi: xoff = xpos - bounds.xhi

        if   ypos <= bounds.ylo: yoff = ypos - bounds.ylo
        elif ypos >= bounds.yhi: yoff = ypos - bounds.yhi

        if xoff != 0 or yoff != 0:
            self.panDisplayBy(xoff, yoff)


    def zoomTo(self, xlo, xhi, ylo, yhi):
        """Zooms the canvas to the given rectangle, specified in
        horizontal/vertical display coordinates.
        """

        # We are going to convert the rectangle specified by
        # the inputs into a zoom value, set the canvas zoom
        # level, and then centre the canvas on the rectangle.

        # Middle of the rectangle, used
        # at the end for centering
        xmid = xlo + (xhi - xlo) / 2.0
        ymid = ylo + (yhi - ylo) / 2.0

        # Size of the rectangle
        rectXlen = abs(xhi - xlo)
        rectYlen = abs(yhi - ylo)

        if rectXlen == 0: return
        if rectYlen == 0: return

        # Size of the overlay bounding
        # box, and the zoom value limits
        opts       = self.opts
        xmin, xmax = self.displayCtx.bounds.getRange(opts.xax)
        ymin, ymax = self.displayCtx.bounds.getRange(opts.yax)
        zoommin    = opts.getAttribute('zoom', 'minval')
        zoommax    = opts.getAttribute('zoom', 'maxval')

        xlen    = xmax    - xmin
        ylen    = ymax    - ymin
        zoomlen = zoommax - zoommin

        # Calculate the ratio of the
        # rectangle to the canvas limits
        xratio = rectXlen / xlen
        yratio = rectYlen / ylen
        ratio  = max(xratio, yratio)

        # Calculate the zoom from this ratio -
        # this is the inverse of the zoom->canvas
        # bounds calculation, as implemented in
        # _applyZoom.
        zoom  = zoommin / ratio
        zoom  = ((zoom - zoommin) / zoomlen) ** (1.0 / 3.0)
        zoom  = zoommin + zoom * zoomlen

        # Update the zoom value, call updateDisplayBounds
        # to apply the new zoom to the display bounds,
        # then centre the display on the calculated
        # centre point.
        with props.skip(opts, 'zoom',          self.name), \
             props.skip(opts, 'displayBounds', self.name):
            opts.zoom = zoom
            self._updateDisplayBounds()
            self.centreDisplayAt(xmid, ymid)

        self.Refresh()


    def resetDisplay(self):
        """Resets the :attr:`zoom` to 100%, and sets the canvas display
        bounds to the overaly bounding box (from the
        :attr:`.DisplayContext.bounds`)
        """

        opts = self.opts

        with props.skip(opts, 'zoom', self.name):
            opts.zoom = 100

        with props.suppress(opts, 'displayBounds'):
            self._updateDisplayBounds()

        self.Refresh()


    def getAnnotations(self):
        """Returns an :class:`.Annotations` instance, which can be used to
        annotate the canvas.
        """
        return self._annotations


    def globjectId(self, overlay):
        """Returns a key that can be used to uniquely identify a
        :class:`.GLObject` for the given overlay. ``GLObject`` instances may be
        shared between different ``SliceCanvas`` instances (specifically, the
        three canvases of an :class:`.OrthoPanel`) using the :mod:`.resources`
        module, using this key.
        """
        return ('GLObject', id(self.displayCtx), overlay)


    def getGLObject(self, overlay):
        """Returns the :class:`.GLObject` associated with the given
        ``overlay``, or ``None`` if there isn't one.
        """
        globj = self._glObjects.get(overlay, None)

        # globjs can be set to False
        if not globj: return None
        else:         return globj


    @property
    def viewport(self):
        """Return the current viewport, as  a sequence of three ``(low, high)``
        values, defining the bounding box in the display coordinate system.

        This method will return ``None`` if :meth:`_draw` has not yet been
        called.
        """
        return self.__viewport


    @property
    def viewMatrix(self):
        """Returns the current model view matrix. """
        return self.__viewMatrix


    @property
    def projectionMatrix(self):
        """Returns the current projection matrix. """
        return self.__projectionMatrix


    @property
    def mvpMatrix(self):
        """Returns the current model*view*projection matrix. """
        return affine.concat(self.__projectionMatrix, self.__viewMatrix)


    def _initGL(self):
        """Call the :meth:`_overlayListChanged` method - it will generate
        any necessary GL data for each of the overlays.
        """
        self._overlayListChanged()


    def _highDpiChange(self, *a):
        """Called when the :attr:`.SliceCanvasOpts.highDpi` property
        changes. Calls the :meth:`.GLCanvasTarget.EnableHighDPI` method.
        """
        self.EnableHighDPI(self.opts.highDpi)


    def _zAxisChanged(self, *a):
        """Called when the :attr:`zax` property is changed. Notifies
        the :class:`.Annotations` instance, and calls :meth:`resetDisplay`.
        """

        opts = self.opts

        log.debug('{}'.format(opts.zax))
        self.resetDisplay()


    def __overlayTypeChanged(self, value, valid, display, name):
        """Called when the :attr:`.Display.overlayType` setting for any
        overlay changes. Makes sure that an appropriate :class:`.GLObject`
        has been created for the overlay (see the :meth:`__genGLObject`
        method).
        """

        log.debug('GLObject representation for {} '
                  'changed to {}'.format(display.name,
                                         display.overlayType))

        self.__regenGLObject(display.overlay)
        self.Refresh()


    def __regenGLObject(self, overlay, refresh=True):
        """Destroys any existing :class:`.GLObject` associated with the given
        ``overlay``, and creates a new one (via the :meth:`__genGLObject`
        method).
        """

        # Tell the previous GLObject (if
        # any) to clean up after itself

        globj = self._glObjects.pop(overlay, None)
        if globj:
            globj.deregister(self.name)
            glresources.delete(self.globjectId(overlay))

        self.__genGLObject(overlay, refresh)


    def __genGLObject(self, overlay, refresh=True):
        """Creates a :class:`.GLObject` instance for the given ``overlay``.
        Does nothing if a ``GLObject`` already exists for the given overlay.

        If ``refresh`` is ``True`` (the default), the :meth:`Refresh` method
        is called after the ``GLObject`` has been created.

        .. note:: If running in ``wx`` (i.e. via a :class:`.WXGLSliceCanvas`),
                  the :class:`.GLObject` instnace will be created on the
                  ``wx.EVT_IDLE`` lopp (via the :mod:`.idle` module).
        """

        display = self.displayCtx.getDisplay(overlay)

        if overlay in self._glObjects:
            return

        # We put a placeholder value in
        # the globjects dictionary, so
        # that the _draw method knows
        # that creation for this overlay
        # is pending.
        self._glObjects[overlay] = False

        def create():

            if not self or self.destroyed:
                return

            # The overlay has been removed from the
            # globjects dictionary between the time
            # the pending flag was set above, and
            # the time that this create() call was
            # executed. Possibly because the overlay
            # was removed between these two events.
            # All is well, just ignore it.
            if overlay not in self._glObjects:
                return

            # We need a GL context to create a new GL
            # object. If we can't get it now, the GL
            # object creation will be re-scheduled on
            # the next call to _draw (via _getGLObjects).
            if not self._setGLContext():

                # Clear the pending flag so
                # this GLObject creation
                # gets re-scheduled.
                self._glObjects.pop(overlay)
                return

            globj = glresources.get(self.globjectId(overlay),
                                    globject.createGLObject,
                                    overlay,
                                    self.overlayList,
                                    self.displayCtx,
                                    False)

            if globj is not None:
                globj.register(self.name, self.__onGLObjectUpdate)

            self._glObjects[overlay] = globj

            display.addListener('overlayType',
                                self.name,
                                self.__overlayTypeChanged,
                                overwrite=True)

            display.addListener('enabled',
                                self.name,
                                self.Refresh,
                                overwrite=True)

            if refresh:
                self.Refresh()

        create = status.reportErrorDecorator(
            strings.titles[  self, 'globjectError'],
            strings.messages[self, 'globjectError'].format(overlay.name))(
                create)

        idle.idle(create)


    def __onGLObjectUpdate(self, globj, *a):
        """Called when a :class:`.GLObject` has been updated, and needs to be
        redrawn.
        """

        # we can sometimes get called after
        # being destroyed (e.g. during testing)
        if self.destroyed:
            return
        self.Refresh()


    def _overlayListChanged(self, *args, **kwargs):
        """This method is called every time an overlay is added or removed
        to/from the overlay list.

        For newly added overlays, calls the :meth:`__genGLObject` method,
        which initialises the OpenGL data necessary to render the
        overlay.
        """

        if self.destroyed:
            return

        # Destroy any GL objects for overlays
        # which are no longer in the list
        for ovl, globj in list(self._glObjects.items()):
            if ovl not in self.overlayList:
                self._glObjects.pop(ovl)
                if globj:
                    glresources.delete(self.globjectId(ovl))

        # Create a GL object for any new overlays,
        # and attach a listener to their display
        # properties so we know when to refresh
        # the canvas.
        for overlay in self.overlayList:

            # A GLObject already exists
            # for this overlay
            if overlay in self._glObjects:
                continue

            self.__regenGLObject(overlay, refresh=False)

        # All the GLObjects are created using
        # idle.idle, so we call refresh in the
        # same way to make sure it gets called
        # after all the GLObject creations.
        def refresh():

            # This SliceCanvas might get
            # destroyed before this idle
            # task is executed
            if not self or self.destroyed:
                return

            self.Refresh()

        idle.idle(refresh)


    def _getGLObjects(self):
        """Called by :meth:`_draw`. Builds a list of all :class:`.GLObjects`
        to be drawn.

        :returns: A list of overlays, and a list of corresponding
                  :class:`.GLObjects` to be drawn.
        """

        overlays = []
        globjs   = []
        for ovl in self.displayCtx.getOrderedOverlays():

            globj = self._glObjects.get(ovl, None)

            # If an overlay does not yet have a corresponding
            # GLObject, we presume that it hasn't been created
            # yet, so we'll tell genGLObject to create one for
            # it.
            if globj is None:
                self.__genGLObject(ovl)

            # If there is a value for this overlay in
            # the globjects dictionary, but it evaluates
            # to False, then GLObject creation has been
            # scheduled for the overlay - see genGLObject.
            elif globj:
                overlays.append(ovl)
                globjs  .append(globj)

        return overlays, globjs


    def _overlayBoundsChanged(self, *args, **kwargs):
        """Called when the :attr:`.DisplayContext.bounds` are changed.
        Initialises/resets the display bounds, and/or preserves the zoom
        level if necessary.

        :arg preserveZoom: Must be passed as a keyword argument. If ``True``
                           (the default), the :attr:`zoom` value is adjusted
                           so that the effective zoom is preserved
        """

        preserveZoom = kwargs.get('preserveZoom', True)

        opts = self.opts
        xax  = opts.xax
        yax  = opts.yax
        xmin = self.displayCtx.bounds.getLo(xax)
        xmax = self.displayCtx.bounds.getHi(xax)
        ymin = self.displayCtx.bounds.getLo(yax)
        ymax = self.displayCtx.bounds.getHi(yax)
        width, height = self.GetSize()

        if np.isclose(xmin, xmax) or width == 0 or height == 0:
            return

        if not preserveZoom or opts.displayBounds.xlen == 0:
            self.resetDisplay()
            return

        # Figure out the scaling factor that
        # would preserve the current zoom
        # level for the new display bounds.
        xmin, xmax, ymin, ymax = glroutines.preserveAspectRatio(
            width, height, xmin, xmax, ymin, ymax)

        scale = opts.displayBounds.xlen / (xmax - xmin)

        # Adjust the zoom value so that the
        # effective zoom stays the same
        with props.suppress(opts, 'zoom'):
            opts.zoom = self.scaleToZoom(scale)


    def _displaySpaceChanged(self, *a):
        """Called when the :attr:`.DisplayContext.displaySpace` changes. Resets
        the display bounds and zoom.
        """
        self.resetDisplay()


    def _zoomChanged(self, *a):
        """Called when the :attr:`zoom` property changes. Updates the
        display bounds.
        """
        opts = self.opts
        loc = [opts.pos[opts.xax], opts.pos[opts.yax]]
        self._updateDisplayBounds(oldLoc=loc)


    def zoomToScale(self, zoom):
        """Converts the given zoom value into a scaling factor that can be
        multiplied by the display bounds width/height.

        Zoom is specified as a percentage.  At 100% the full scene takes up
        the full display.

        In order to make the zoom smoother at low levels, we re-scale the zoom
        value to be exponential across the range.

        This is done by transforming the zoom from ``[zmin, zmax]`` into
        ``[0.0, 1.0]``, then turning it from linear ``[0.0, 1.0]`` to
        exponential ``[0.0, 1.0]``, and then finally transforming it back to
        ``[zmin - zmax]``.

        However there is a slight hack in that, if the zoom value is less than
        100%, it will be applied linearly (i.e. 50% will cause the width/height
        to be scaled by 50%).
        """

        # Assuming that minval == 100.0
        opts = self.opts
        zmin = opts.getAttribute('zoom', 'minval')
        zmax = opts.getAttribute('zoom', 'maxval')
        zlen = zmax - zmin

        # Don't break the maths below
        if zoom <= 0:
            zoom = 1

        # Normal behaviour
        if zoom >= 100:

            # [100 - zmax] -> [0.0 - 1.0] -> exponentify -> [100 - zmax]
            zoom = (zoom - zmin)      / zlen
            zoom = zmin + (zoom ** 3) * zlen

            # Then we transform the zoom from
            # [100 - zmax] to [1.0 - 0.0] -
            # this value is used to scale the
            # bounds.
            scale = zmin / zoom

        # Hack for zoom < 100
        else:
            scale = 100.0 / zoom

        return scale


    def scaleToZoom(self, scale):
        """Converts the given zoom scaling factor into a zoom percentage.
        This method performs the reverse operation to the :meth:`zoomToScale`
        method.
        """

        opts = self.opts
        zmin = opts.getAttribute('zoom', 'minval')
        zmax = opts.getAttribute('zoom', 'maxval')
        zlen = zmax - zmin

        if scale > 1:
            zoom = 100.0 / scale

        else:

            # [100 - zmax] -> [0.0 - 1.0] -> de-exponentify -> [100 - zmax]
            zoom = zmin / scale
            zoom = (zoom - zmin) / zlen
            zoom = np.power(zoom, 1.0 / 3.0)
            zoom = zmin + zoom * zlen

        return zoom


    def _applyZoom(self, xmin, xmax, ymin, ymax):
        """*Zooms* in to the given rectangle according to the current value
        of the zoom property Returns a 4-tuple containing the updated bound
        values.
        """

        zoomFactor = self.zoomToScale(self.opts.zoom)

        xlen    = xmax - xmin
        ylen    = ymax - ymin
        newxlen = xlen * zoomFactor
        newylen = ylen * zoomFactor

        # centre the zoomed-in rectangle
        # on the provided limits
        xmid = xmin + 0.5 * xlen
        ymid = ymin + 0.5 * ylen

        # new x/y min/max bounds
        xmin = xmid - 0.5 * newxlen
        xmax = xmid + 0.5 * newxlen
        ymin = ymid - 0.5 * newylen
        ymax = ymid + 0.5 * newylen

        return (xmin, xmax, ymin, ymax)


    def _updateDisplayBounds(self, bbox=None, oldLoc=None):
        """Called on canvas resizes, overlay bound changes, and zoom changes.

        Calculates the bounding box, in display coordinates, to be displayed
        on the canvas. Stores this bounding box in the :attr:`displayBounds`
        property. If any of the parameters are not provided, the
        :attr:`.DisplayContext.bounds` are used.


        .. note:: This method is used internally, and also by the
                  :class:`.WXGLSliceCanvas` class.

        .. warning:: This code assumes that, if the display coordinate system
                     has changed, the display context location has already
                     been updated.  See the
                     :meth:`.DisplayContext.__displaySpaceChanged` method.


        :arg bbox:   Tuple containing four values:

                      - Minimum x (horizontal) value to be in the display
                        bounds.

                      - Maximum x value to be in the display bounds.

                      - Minimum y (vertical) value to be in the display bounds.

                      - Maximum y value to be in the display bounds.

        :arg oldLoc: If provided, should be the ``(x, y)`` location shown on
                     this ``SliceCanvas`` - the new display bounds will be
                     adjusted so that this location remains the same, with
                     respect to the new field of view.
        """

        opts = self.opts

        if bbox is None:
            bbox = (self.displayCtx.bounds.getLo(opts.xax),
                    self.displayCtx.bounds.getHi(opts.xax),
                    self.displayCtx.bounds.getLo(opts.yax),
                    self.displayCtx.bounds.getHi(opts.yax))

        xmin = bbox[0]
        xmax = bbox[1]
        ymin = bbox[2]
        ymax = bbox[3]

        # Save the display bounds in case
        # we need to preserve them with
        # respect to the current display
        # location.
        width, height                      = self.GetSize()
        oldxmin, oldxmax, oldymin, oldymax = opts.displayBounds[:]

        log.debug('{}: Required display bounds: '
                  'X: ({: 5.1f}, {: 5.1f}) Y: ({: 5.1f}, {: 5.1f})'.format(
                      opts.zax, xmin, xmax, ymin, ymax))

        # Adjust the bounds to preserve the
        # x/y aspect ratio, and according to
        # the current zoom level.
        xmin, xmax, ymin, ymax = glroutines.preserveAspectRatio(
            width, height, xmin, xmax, ymin, ymax)
        xmin, xmax, ymin, ymax = self._applyZoom(xmin, xmax, ymin, ymax)

        # If a location (oldLoc) has been provided,
        # adjust the bounds so they are consistent
        # with respect to that location.
        if oldLoc and (oldxmax > oldxmin) and (oldymax > oldymin):

            # Calculate the normalised distance from the
            # old cursor location to the old bound corner
            oldxoff = (oldLoc[0] - oldxmin) / (oldxmax - oldxmin)
            oldyoff = (oldLoc[1] - oldymin) / (oldymax - oldymin)

            # Re-set the new bounds to the current
            # display location, offset by the same
            # amount that it used to be (as
            # calculated above).
            #
            # N.B. This code assumes that, if the display
            #      coordinate system has changed, the display
            #      context location has already been updated.
            #      See the DisplayContext.__displaySpaceChanged
            #      method.
            xloc = opts.pos[opts.xax]
            yloc = opts.pos[opts.yax]

            xlen = xmax - xmin
            ylen = ymax - ymin

            xmin = xloc - oldxoff * xlen
            ymin = yloc - oldyoff * ylen

            xmax = xmin + xlen
            ymax = ymin + ylen

        log.debug('{}: Final display bounds: '
                  'X: ({: 5.1f}, {: 5.1f}) Y: ({: 5.1f}, {: 5.1f})'.format(
                      opts.zax, xmin, xmax, ymin, ymax))

        opts.displayBounds[:] = (xmin, xmax, ymin, ymax)


    def _setViewport(self, invertX=None, invertY=None):
        """Calculates the GL bounds, projection, and model view matrices. They
        are stored as attributes which are accessible via the
        :meth:`viewport`, :meth:`projectionMatrix` and :meth:`viewMatrix`
        methods.

        :arg invertX: Invert the X axis. If not provided, taken from
                      :attr:`invertX`.

        :arg invertY: Invert the Y axis. If not provided, taken from
                      :attr:`invertY`.
        """

        opts          = self.opts
        xax           = opts.xax
        yax           = opts.yax
        zax           = opts.zax
        xmin          = opts.displayBounds.xlo
        xmax          = opts.displayBounds.xhi
        ymin          = opts.displayBounds.ylo
        ymax          = opts.displayBounds.yhi
        zmin          = self.displayCtx.bounds.getLo(zax)
        zmax          = self.displayCtx.bounds.getHi(zax)

        if invertX is None: invertX = opts.invertX
        if invertY is None: invertY = opts.invertY

        # If there is  no space to draw, do nothing
        if (xmin == xmax) or (ymin == ymax):
            self.__viewport         = None
            self.__projectionMatrix = None
            self.__viewMatrix       = None
            return

        # Add a bit of padding to the depth limits
        zmin -= 1e-3
        zmax += 1e-3

        lo = [None] * 3
        hi = [None] * 3

        lo[xax], hi[xax] = xmin, xmax
        lo[yax], hi[yax] = ymin, ymax
        lo[zax], hi[zax] = zmin, zmax

        # calculate projection and mv
        # matrices for 2D ortho
        projmat, mvmat = glroutines.show2D(
            xax, yax, lo, hi, invertX, invertY)

        # store a copy of the final bounds and
        # proj/mv matrices interested parties can
        # retrieve them via the viewport/
        # projectionMatrix/ viewMatrix methods.
        self.__viewport         = [(lo[0], hi[0]), (lo[1], hi[1]), (lo[2], hi[2])]
        self.__projectionMatrix = projmat
        self.__viewMatrix       = mvmat


    def _drawCursor(self):
        """Draws a green cursor at the current X/Y position."""

        copts      = self.opts
        ovl        = self.displayCtx.getSelectedOverlay()
        xmin, xmax = self.displayCtx.bounds.getRange(copts.xax)
        ymin, ymax = self.displayCtx.bounds.getRange(copts.yax)
        x          = copts.pos[copts.xax]
        y          = copts.pos[copts.yax]
        lines      = []

        # Just show a vertical line at xpos,
        # and a horizontal line at ypos
        if not copts.cursorGap:
            lines.append((x,    ymin, x,    ymax))
            lines.append((xmin, y,    xmax, y))

        # Draw vertical/horizontal cursor lines,
        # with a gap at the cursor centre

        # Not a NIFTI image - just
        # use a fixed gap size
        elif ovl is None or not isinstance(ovl, fslimage.Nifti):

            lines.append((xmin,    y,       x - 0.5, y))
            lines.append((x + 0.5, y,       xmax,    y))
            lines.append((x,       ymin,    x,       y - 0.5))
            lines.append((x,       y + 0.5, x,       ymax))

        # If the current overlay is NIFTI, make
        # the gap size match its voxel size
        else:

            # Get the current voxel
            # coordinates,
            dopts = self.displayCtx.getOpts(ovl)
            vox   = dopts.getVoxel(vround=False)

            # Out of bounds of the current
            # overlay, fall back to using
            # a fixed size gap
            if vox is None:
                xlow  = x - 0.5
                xhigh = x + 0.5
                ylow  = y - 0.5
                yhigh = y + 0.5
            else:

                vox  = np.array(vox, dtype=np.float32)

                # Figure out the voxel coord axes
                # that (approximately) correspond
                # with the display x/y axes.
                axes = ovl.axisMapping(dopts.getTransform('voxel', 'display'))
                axes = np.abs(axes) - 1
                xax  = axes[copts.xax]
                yax  = axes[copts.yax]

                # Clamp the voxel x/y coords to
                # the voxel edge (round, then
                # offset by 0.5 - integer coords
                # correspond to the voxel centre).
                vox[xax] = np.round(vox[xax]) - 0.5
                vox[yax] = np.round(vox[yax]) - 0.5

                # Get the voxels that are above
                # and next to our current voxel.
                voxx = np.copy(vox)
                voxy = np.copy(vox)

                voxx[xax] += 1
                voxy[yax] += 1

                # Transform those integer coords back
                # into display coordinates to get the
                # display location on the voxel boundary.
                vloc  = dopts.transformCoords(vox,  'voxel', 'display')
                vlocx = dopts.transformCoords(voxx, 'voxel', 'display')
                vlocy = dopts.transformCoords(voxy, 'voxel', 'display')

                xlow  = min(vloc[copts.xax], vlocx[copts.xax])
                xhigh = max(vloc[copts.xax], vlocx[copts.xax])
                ylow  = min(vloc[copts.yax], vlocy[copts.yax])
                yhigh = max(vloc[copts.yax], vlocy[copts.yax])

            lines.append((xmin,  y,     xlow, y))
            lines.append((xhigh, y,     xmax, y))
            lines.append((x,     ymin,  x,    ylow))
            lines.append((x,     yhigh, x,    ymax))

        kwargs = {
            'colour'     : copts.cursorColour,
            'lineWidth'  : copts.cursorWidth
        }

        for line in lines:
            self._annotations.line(*line, **kwargs)
            self._annotations.line(*line, **kwargs)


    def _draw(self, *a):
        """Draws the current scene to the canvas. """

        if self.destroyed:
            return

        width, height = self.GetScaledSize()
        copts         = self.opts
        zpos          = copts.pos[copts.zax]
        axes          = (copts.xax, copts.yax, copts.zax)

        if width == 0 or height == 0:
            return

        if not self._setGLContext():
            return

        gl.glViewport(0, 0, width, height)
        glroutines.clear(copts.bgColour)

        overlays, globjs = self._getGLObjects()

        if len(overlays) == 0:
            return

        # Calculate viewport bounds and
        # projection/modelview matrices
        self._setViewport()
        if self.projectionMatrix is None:
            return

        # Do not draw anything if some globjects
        # are not ready. This is because, if a
        # GLObject was drawn, but is now temporarily
        # not ready (e.g. it has an image texture
        # that is being asynchronously refreshed),
        # drawing the scene now would cause
        # flickering of that GLObject.
        if len(globjs) == 0 or any(not g.ready() for g in globjs):
            return

        for overlay, globj in zip(overlays, globjs):

            display = self.displayCtx.getDisplay(overlay)
            if not display.enabled:
                continue

            log.debug('Drawing %u slice for overlay %s',
                      copts.zax, display.name)

            globj.preDraw()
            globj.draw2D(self, zpos, axes)
            globj.postDraw()

        if copts.showCursor:
            self._drawCursor()

        self._annotations.draw2D(zpos, axes)



#!/usr/bin/env python
#
# gltractogram.py - Logic for rendering GLTractogram instances.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

"""This module contains logic for rendering :class:`.Tractogram` overlays.
"""


import itertools as it

import numpy     as np
import OpenGL.GL as gl

import fsl.utils.idle       as idle
import fsl.transform.affine as affine
import fsl.data.image       as fslimage

import fsleyes.gl           as fslgl
import fsleyes.gl.textures  as textures
import fsleyes.gl.globject  as globject


class GLTractogram(globject.GLObject):
    """The GLTractogram contains logic for drawing :class:`.Tractogram`
    overlays.
    """

    def __init__(self, overlay, overlayList, displayCtx, threedee):
        """Create a :meth:`GLTractogram`. """
        globject.GLObject.__init__(
            self, overlay, overlayList, displayCtx, threedee)

        # Shaders are created in compileShaders.
        # imageTexture created in refreshImageTexture
        #
        # Three separate shader types are used:
        #  - 'orientation' - coloured by streamline orientation
        #  - 'vertexData'  - coloured by per-vertex/streamline data
        #  - 'imageData'   - coloured by data from an Image
        #
        # Separate shader programs are used depending
        # on the TractogramOpts.clipMode -
        #  - 'none'       - no clipping, or clipping by the same
        #                   data used for colouring
        #  - 'vertexData' - clipping by a separate vertex data set
        #  - 'imageData'  - clipping by a separate image data set
        self.shaders        = {
            'orientation' : {'none' : [], 'vertexData' : [], 'imageData' : []},
            'vertexData'  : {'none' : [], 'vertexData' : [], 'imageData' : []},
            'imageData'   : {'none' : [], 'vertexData' : [], 'imageData' : []}}
        self.cmapTexture    = textures.ColourMapTexture(self.name)
        self.negCmapTexture = textures.ColourMapTexture(self.name)
        self.imageTextures  = textures.AuxImageTextureManager(
            self, colour=None, clip=None)

        self.compileShaders()
        self.updateStreamlineData()
        self.refreshImageTexture('clip')
        self.refreshImageTexture('colour')
        self.refreshCmapTextures()
        self.updateColourData(refresh=False)
        self.updateClipData(refresh=False)
        self.addListeners()

        # Call updateShaderState asynchronously,
        # as it may depend on the image textures
        # being ready, which might be prepared
        # off the main thread.
        if 'imageData' not in (self.opts.colourMode, self.opts.clipMode):
            self.updateShaderState()
        else:
            idle.idleWhen(self.updateShaderState, self.ready)


    def addListeners(self):
        """Called by :meth:`__init__`. Adds a bunch of property listeners
        to the :class:`.Display` and :class:`.TractogramOpts` instances.
        """
        opts    = self.opts
        display = self.display
        name    = self.name

        def refresh(*_):
            self.notify()

        def data(*_):
            self.updateStreamlineData()
            self.notify()

        def shader(*_):
            self.updateShaderState()
            self.notify()

        def cmaps(*_):
            self.refreshCmapTextures()
            self.updateShaderState()
            self.notify()

        def colour(*_):
            self.updateColourData()
            self.notifyWhen(self.ready)

        def clip(*_):
            self.updateClipData()
            self.notifyWhen(self.ready)

        opts   .addListener('resolution',          name, shader,  weak=False)
        opts   .addListener('subsample',           name, data,    weak=False)
        opts   .addListener('colourMode',          name, colour,  weak=False)
        opts   .addListener('clipMode',            name, clip,    weak=False)
        opts   .addListener('lineWidth',           name, refresh, weak=False)
        opts   .addListener('xColour',             name, shader,  weak=False)
        opts   .addListener('yColour',             name, shader,  weak=False)
        opts   .addListener('zColour',             name, shader,  weak=False)
        opts   .addListener('suppressX',           name, shader,  weak=False)
        opts   .addListener('suppressY',           name, shader,  weak=False)
        opts   .addListener('suppressZ',           name, shader,  weak=False)
        opts   .addListener('suppressMode',        name, shader,  weak=False)
        opts   .addListener('displayRange',        name, cmaps,   weak=False)
        opts   .addListener('clippingRange',       name, shader,  weak=False)
        opts   .addListener('invertClipping',      name, shader,  weak=False)
        opts   .addListener('cmap',                name, cmaps,   weak=False)
        opts   .addListener('negativeCmap',        name, cmaps,   weak=False)
        opts   .addListener('useNegativeCmap',     name, shader,  weak=False)
        opts   .addListener('gamma',               name, cmaps,   weak=False)
        opts   .addListener('logScale',            name, cmaps,   weak=False)
        opts   .addListener('cmapResolution',      name, cmaps,   weak=False)
        opts   .addListener('interpolateCmaps',    name, cmaps,   weak=False)
        opts   .addListener('invert',              name, cmaps,   weak=False)
        opts   .addListener('modulateAlpha',       name, shader,  weak=False)
        opts   .addListener('invertModulateAlpha', name, shader,  weak=False)
        opts   .addListener('modulateRange',       name, shader,  weak=False)
        display.addListener('alpha',               name, cmaps,   weak=False)


    def removeListeners(self):
        """Called by :meth:`destroy`. Removes all property listeners. """
        opts    = self.opts
        display = self.display
        name    = self.name

        opts   .removeListener('resolution',          name)
        opts   .removeListener('subsample',           name)
        opts   .removeListener('lineWidth',           name)
        opts   .removeListener('colourMode',          name)
        opts   .removeListener('clipMode',            name)
        opts   .removeListener('xColour',             name)
        opts   .removeListener('yColour',             name)
        opts   .removeListener('zColour',             name)
        opts   .removeListener('suppressX',           name)
        opts   .removeListener('suppressY',           name)
        opts   .removeListener('suppressZ',           name)
        opts   .removeListener('suppressMode',        name)
        opts   .removeListener('displayRange',        name)
        opts   .removeListener('clippingRange',       name)
        opts   .removeListener('invertClipping',      name)
        opts   .removeListener('cmap',                name)
        opts   .removeListener('negativeCmap',        name)
        opts   .removeListener('useNegativeCmap',     name)
        opts   .removeListener('gamma',               name)
        opts   .removeListener('logScale',            name)
        opts   .removeListener('cmapResolution',      name)
        opts   .removeListener('interpolateCmaps',    name)
        opts   .removeListener('invert',              name)
        opts   .removeListener('modulateAlpha',       name)
        opts   .removeListener('invertModulateAlpha', name)
        opts   .removeListener('modulateRange',       name)
        display.removeListener('alpha',               name)


    def destroy(self):
        """Removes listeners and destroys textures and shader programs. """

        if self.cmapTexture is not None:
            self.cmapTexture.destroy()

        if self.negCmapTexture is not None:
            self.negCmapTexture.destroy()

        if self.imageTextures is not None:
            self.imageTextures.destroy()

        if self.shaders is not None:
            for shader in self.iterShaders():
                shader.destroy()

        self.cmapTexture    = None
        self.negCmapTexture = None
        self.imageTextures  = None
        self.shaders        = None

        self.removeListeners()
        globject.GLObject.destroy(self)


    def iterShaders(self, colourModes=None, clipModes=None):
        """Returns all shader programs for the specified colour/clipping
        modes.
        """
        if isinstance(colourModes, str):
            colourModes = [colourModes]
        if isinstance(clipModes, str):
            clipModes = [clipModes]
        if colourModes is None or len(colourModes) == 0:
            colourModes = ['orientation', 'vertexData', 'imageData']
        if clipModes is None or len(clipModes) == 0:
            clipModes = ['none', 'vertexData', 'imageData']
        shaders = [self.shaders[m] for m in colourModes]
        shaders = it.chain(*[[s[m] for m in clipModes] for s in shaders])
        return it.chain(*shaders)


    @property
    def destroyed(self):
        """Returns ``True`` if :meth:`destroy` has been called. """
        return self.shaders is None


    def ready(self):
        """Overrides :meth:`.GLObject.ready`. Returns ``True`` if the
        :attr:`.TractogramOpts.colourImage` is ready (or unset).
        """
        return self.imageTextures.texturesReady()


    def getDisplayBounds(self):
        """Overrides :meth:`.GLObject.getDisplayBounds`. Returns a bounding
        box which contains the mesh vertices.
        """
        return (self.opts.bounds.getLo(),
                self.opts.bounds.getHi())


    def normalisedLineWidth(self, canvas, mvp):
        """Returns :attr:`lineWidth`, scaled so that it is with respect to
        normalised device coordinates. Streamline lines/tubes (in 3D) and
        vertices (in 2D) are drawn such that the width/radius is fixed w.r.t.
        the world coordinate system - i.e. for a given ``lineWidth``,
        lines/tubes/circles will be small when zoomed out, and big when
        zoomed in.

        :arg canvas: Canvas being drawn on
        :arg mvp:    Current MVP matrix
        """
        # Line width is fixed at 0.1 in world
        # units. Below we convert it so it is
        # in terms of NDCs.
        lineWidth =  self.opts.lineWidth / 10

        if self.threedee:
            # We don't apply the scene3d rotation, because
            # the projection matrix adds an uneven scaling
            # to the depth axis (see routines.ortho3D),
            # which will affect scaling when rotated to be
            # in line with that axis.  I may revisit this in
            # the future, as the ortho3D function is a bit
            # nuts in how it handles depth (caused my my
            # lack of understanding of near/far clipping).
            scaling   = affine.concat(canvas.projectionMatrix,
                                      canvas.viewScale)
            lineWidth = [lineWidth * scaling[0, 0],
                         lineWidth * scaling[1, 1]]

        else:
            # return separate scales for each axis
            lineWidth = affine.transform([lineWidth] * 3, mvp, vector=True)

        return lineWidth


    @property
    def clipImageTexture(self):
        """Return a reference to an :class:`.ImageTexture` which contains
        data for clipping, when :attr:`.TractogramOpts.clipMode` is set
        to an :class:`.Image`.
        """
        return self.imageTextures.texture('clip')


    @property
    def colourImageTexture(self):
        """Return a reference to an :class:`.ImageTexture` which contains
        data for colouring, when :attr:`.TractogramOpts.colourMode` is set
        to an :class:`.Image`.
        """
        return self.imageTextures.texture('colour')


    def shaderAttributeArgs(self):
        """Returns keyword arguments to pass to :meth:`.GLSLShader.setAtt`
        for all per-vertex attributes. The GL21 2D rendering logic uses
        instanced rendering, so a divisor must be set for all vertex
        attributes.
        """
        if fslgl.GL_COMPATIBILITY != '2.1': return {}
        if self.threedee:                   return {}
        else:                               return {'divisor' : 1}


    def compileShaders(self):
        """Called by :meth:`__init__`. Calls
        :func:`.gl21.gltractogram_funcs.compileShaders` or
        :func:`.gl33.gltractogram_funcs.compileShaders`, passes vertex data to
        the shader programs, and calls :meth:`updateShaderState`
        """
        fslgl.gltractogram_funcs.compileShaders(self)


    def updateStreamlineData(self):
        """Prepares streamline data and passes it to GL. This method is called
        on creation, and whenever the :attr:`.TractogramOpts.subsample` setting
        changes.

        For 2D views (ortho/lightbox), the tractogram is drawn as ``GL_POINT``
        points using glDrawArrays (or equivalent). For 3D views, streamlines
        are drawn as ``GL_LINE_STRIP`` lines using offsets/counts into the
        vertex array.
        """

        ovl         = self.overlay
        opts        = self.opts
        subsamp     = opts.subsample
        nverts      = ovl.nvertices
        nstrms      = ovl.nstreamlines
        kwargs      = self.shaderAttributeArgs()
        threedee    = self.threedee
        indices     = None

        # randomly select a subset of streamlines
        # (3D) or vertices (2D).
        if subsamp < 100:
            if threedee: lim = nstrms
            else:        lim = nverts
            n       = int(lim * subsamp / 100)
            indices = np.random.choice(lim, n)
            indices = np.sort(indices)

            if threedee:
                vertices, offsets, counts, indices = ovl.subset(indices)
                orients = ovl.vertexOrientations[indices]
            else:
                offsets, counts = [0], [0]
                vertices        = ovl.vertices[          indices]
                orients         = ovl.vertexOrientations[indices]
        else:
            vertices = ovl.vertices
            orients  = ovl.vertexOrientations
            offsets  = ovl.offsets
            counts   = ovl.lengths

        # Orientation is used for RGB colouring.
        # We have to apply abs so that GL doesn't
        # interpolate across -ve/+ve boundaries.
        # when passing from vertex shader through
        # to fragment shader.
        self.vertices =        np.asarray(vertices, dtype=np.float32)
        self.orients  = np.abs(np.asarray(orients,  dtype=np.float32))
        self.offsets  =        np.asarray(offsets,  dtype=np.int32)
        self.counts   =        np.asarray(counts,   dtype=np.int32)
        self.indices  = indices

        # upload vertices/orients/indices to GL.
        # For 3D, offsets/counts are passed on
        # each draw
        for shader in self.iterShaders('orientation'):
            with shader.loaded():
                shader.setAtt('vertex', self.vertices, **kwargs)
                shader.setAtt('orient', self.orients,  **kwargs)
        for shader in self.iterShaders(('vertexData', 'imageData')):
            with shader.loaded():
                shader.setAtt('vertex', self.vertices, **kwargs)
        if opts.effectiveColourMode == 'vertexData':
            self.updateColourData()
        if opts.effectiveClipMode == 'vertexData':
            self.updateClipData()


    def updateShaderState(self):
        """Passes display properties as uniform values to the shader programs.
        """
        opts                = self.opts
        display             = self.display
        colours, xform      = opts.getVectorColours()
        colourScale         = xform[0, 0]
        colourOffset        = xform[0, 3]
        cmapXform           = self.cmapTexture.getCoordinateTransform()
        cmapScale           = cmapXform[0, 0]
        cmapOffset          = cmapXform[0, 3]
        modScale, modOffset = opts.modulateScaleOffset()
        sameData            = opts.clipMode in (None, opts.colourMode)

        # We scale alpha exponentially as, for a
        # typical tractogram with many streamlines,
        # transparency only starts to take effect
        # when display.alpha ~= 20. Also, when
        # alpha < 100, we draw the tractogram twice
        # - see the gltractogram_funcs.draw3D
        # function for more details.
        if display.alpha < 100:
            alpha         = (display.alpha / 100) ** 2
            colours[0][3] = alpha
            colours[0][3] = alpha
            colours[1][3] = alpha
            colours[2][3] = alpha

        for shader in self.iterShaders('orientation'):
            with shader.loaded():
                shader.set('xColour',      colours[0])
                shader.set('yColour',      colours[1])
                shader.set('zColour',      colours[2])
                shader.set('colourScale',  colourScale)
                shader.set('colourOffset', colourOffset)
                shader.set('resolution',   opts.resolution)
                shader.set('lighting',     False)

        for shader in self.iterShaders('orientation',
                                       ['vertexData', 'imageData']):
            with shader.loaded():
                shader.set('invertClip',   opts.invertClipping)
                shader.set('clipLow',      opts.clippingRange.xlo)
                shader.set('clipHigh',     opts.clippingRange.xhi)

        for shader in self.iterShaders(['vertexData', 'imageData']):
            with shader.loaded():
                shader.set('resolution',    opts.resolution)
                shader.set('cmap',          0)
                shader.set('negCmap',       1)
                shader.set('useNegCmap',    opts.useNegativeCmap)
                shader.set('cmapScale',     cmapScale)
                shader.set('cmapOffset',    cmapOffset)
                shader.set('sameData',      sameData)
                shader.set('invertClip',    opts.invertClipping)
                shader.set('clipLow',       opts.clippingRange.xlo)
                shader.set('clipHigh',      opts.clippingRange.xhi)
                shader.set('modulateAlpha', opts.modulateAlpha)
                shader.set('modScale',      modScale)
                shader.set('modOffset',     modOffset)
                shader.set('lighting',      False)


    def updateColourData(self, refresh=True):
        """Called when :class:`.TractogramOpts.colourMode` changes. Passes
        data to the shader programs.
        """

        opts    = self.opts
        ovl     = self.overlay
        indices = self.indices
        cmode   = opts.effectiveColourMode
        kwargs  = self.shaderAttributeArgs()

        if cmode == 'orientation':
            return

        if cmode == 'vertexData':
            data = ovl.getVertexData(opts.colourMode)
            if indices is not None:
                data = data[indices]
            for shader in self.iterShaders('vertexData'):
                with shader.loaded():
                    shader.setAtt('vertexData', data, **kwargs)

        if refresh and cmode == 'imageData':
            self.refreshImageTexture('colour')


    def updateClipData(self, refresh=True):
        """Called when :class:`.TractogramOpts.clipMode` changes. Passes
        data to the shader programs.
        """

        opts    = self.opts
        ovl     = self.overlay
        indices = self.indices
        cmode   = opts.effectiveClipMode
        kwargs  = self.shaderAttributeArgs()

        if cmode == 'none':
            return

        if cmode == 'vertexData':
            data = ovl.getVertexData(opts.clipMode)
            if indices is not None:
                data = data[indices]
            for shader in self.iterShaders(None, 'vertexData'):
                with shader.loaded():
                    shader.setAtt('clipVertexData', data, **kwargs)

        elif refresh and cmode == 'imageData':
            self.refreshImageTexture('clip')


    def refreshImageTexture(self, which):
        """Called on changes to :attr:`.TractogramOpts.colourMode` and
        :attr:`.TractogramOpts.clipMode`. Refreshes the :class:`.ImageTexture`
        objects as needed.
        """

        opts = self.opts

        if which == 'colour': image = opts.colourMode
        else:                 image = opts.clipMode

        # Not currently colouring/clipping by image.
        if not isinstance(image, fslimage.Image):
            return

        # image already registered
        if image is self.imageTextures.image(which):
            return

        # When the texture has been prepared,
        # we need to tell the shader programs
        # how to use it.
        if which == 'colour': callback = self.colourImageTextureChanged
        else:                 callback = self.clipImageTextureChanged

        self.imageTextures.registerAuxImage(
            which, image, callback=callback, notify=True)
        self.imageTextures.texture(which).register(self.name, callback)


    def colourImageTextureChanged(self, *_):
        """Calls :meth:`imageTextureChanged`. """
        self.imageTextureChanged('colour')


    def clipImageTextureChanged(self, *_):
        """Calls :meth:`imageTextureChanged`. """
        self.imageTextureChanged('clip')


    def imageTextureChanged(self, which):
        """Called when :attr:`.TractogramOpts.colourMode` or
        :attr:`.TractogramOpts.clipMode` is set to an image, and the
        underlying :class:`.ImageTexture` changes. Sets some shader uniforms
        accordingly.
        """

        if which == 'colour': image = self.opts.colourMode
        else:                 image = self.opts.clipMode

        texture   = self.imageTextures.texture(which)
        opts      = self.displayCtx.getOpts(image)
        w2tXform  = opts.getTransform('world', 'texture')
        voxXform  = texture.voxValXform
        voxScale  = voxXform[0, 0]
        voxOffset = voxXform[0, 3]

        if which == 'colour':
            for shader in self.iterShaders('imageData'):
                with shader.loaded():
                    shader.set('imageTexture',  2)
                    shader.set('texCoordXform', w2tXform)
                    shader.set('voxScale',      voxScale)
                    shader.set('voxOffset',     voxOffset)
        elif which == 'clip':
            for shader in self.iterShaders([], ['imageData']):
                with shader.loaded():
                    shader.set('clipTexture',       3)
                    shader.set('clipTexCoordXform', w2tXform)
                    shader.set('clipValScale',      voxScale)
                    shader.set('clipValOffset',     voxOffset)


    def refreshCmapTextures(self):
        """Called when various :class:`.Display` or :class:`.TractogramOpts``
        properties change. Refreshes the :class:`.ColourMapTexture` instances
        corresponding to the :attr:`.TractogramOpts.cmap` and
        :attr:`.TractogramOpts.negativeCmap` properties.
        """

        display  = self.display
        opts     = self.opts
        alpha    = display.alpha / 100.0
        cmap     = opts.cmap
        interp   = opts.interpolateCmaps
        res      = opts.cmapResolution
        negCmap  = opts.negativeCmap
        gamma    = opts.realGamma(opts.gamma)
        logScale = opts.logScale
        invert   = opts.invert
        dmin     = opts.displayRange[0]
        dmax     = opts.displayRange[1]

        if interp: interp = gl.GL_LINEAR
        else:      interp = gl.GL_NEAREST

        # Scale alpha exponentially (see inline
        # comments in updateShaderState)
        if alpha < 1:
            alpha = alpha ** 2

        self.cmapTexture.set(cmap=cmap,
                             invert=invert,
                             alpha=alpha,
                             resolution=res,
                             gamma=gamma,
                             logScale=logScale,
                             interp=interp,
                             displayRange=(dmin, dmax))

        self.negCmapTexture.set(cmap=negCmap,
                                invert=invert,
                                alpha=alpha,
                                resolution=res,
                                gamma=gamma,
                                logScale=logScale,
                                interp=interp,
                                displayRange=(dmin, dmax))


    def preDraw(self):
        """Called before :meth:`draw2D`/:meth:`draw3D`. Binds textures as
        needed.
        """

        colourMode = self.opts.effectiveColourMode
        clipMode   = self.opts.effectiveClipMode

        cmaps     = colourMode in ('imageData', 'vertexData')
        colourTex = colourMode == 'imageData'
        clipTex   = clipMode   == 'imageData'

        if cmaps:
            self.cmapTexture   .bindTexture(gl.GL_TEXTURE0)
            self.negCmapTexture.bindTexture(gl.GL_TEXTURE1)
        if colourTex:
            self.colourImageTexture.bindTexture(gl.GL_TEXTURE2)
        if clipTex:
            self.clipImageTexture.bindTexture(gl.GL_TEXTURE3)


    def draw2D(self, canvas, zpos, axes, xform=None):
        """Draws a 2D slice through the tractogram. Calls
        :func:`.gl21.gltractogram_funcs.draw2D` or
        :func:`.gl33.gltractogram_funcs.draw2D`.
        """

        if xform is None:
            xform = np.eye(4)

        opts = self.opts
        zax  = axes[2]

        # We draw a 2D slice through the tractogram by
        # manipulating the projection matrix so that z
        # coordinates within the slice are mapped to
        # the range (-1, +1). Vertices with z outside
        # of that range will be clipped by GL.
        projmat       = np.array(canvas.projectionMatrix)
        step          = opts.sliceWidth(zax)
        zlo           = zpos - step
        zhi           = zpos + step
        projmat[2, 2] = 2 / (zhi - zlo)
        projmat[2, 3] = -(zhi + zlo) / (zhi - zlo)

        # The routines.show2D function encodes a
        # -ve scale on the yaxis in the view
        # matrix.  We need to accommodate it
        # here.
        if zax == 1:
            projmat[2, 2] *= -1

        viewmat   = canvas.viewMatrix
        strm2disp = opts.displayTransform
        mvp       = affine.concat(projmat, viewmat, xform, strm2disp)

        fslgl.gltractogram_funcs.draw2D(self, canvas, mvp)


    def draw3D(self, *args, **kwargs):
        """Calls :func:`.gl21.gltractogram_funcs.draw3D` or
        :func:`.gl33.gltractogram_funcs.draw3D`.
        """
        fslgl.gltractogram_funcs.draw3D(self, *args, **kwargs)


    def postDraw(self):
        """Called after :meth:`draw2D`/:meth:`draw3D`. Unbinds textures as
        needed.
        """

        colourMode = self.opts.effectiveColourMode
        clipMode   = self.opts.effectiveClipMode

        cmaps     = colourMode in ('imageData', 'vertexData')
        colourTex = colourMode == 'imageData'
        clipTex   = clipMode   == 'imageData'

        if cmaps:
            self.cmapTexture   .unbindTexture()
            self.negCmapTexture.unbindTexture()
        if colourTex:
            self.colourImageTexture.unbindTexture()
        if clipTex:
            self.clipImageTexture.unbindTexture()



#!/usr/bin/env python
#
# orthocropprofile.py - The OrthoCropProfile class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoCropProfile` class, an interaction
:class:`.Profile` for :class:`.OrthoPanel` views, which is used by the
:class:`.CropImagePanel`.
"""


import logging

import          wx
import numpy as np

import fsl.data.image                     as fslimage
import fsl.utils.run                      as fslrun
import fsl.utils.tempdir                  as tempdir
from   fsl.utils.platform import platform as fslplatform
import fsleyes_props                      as props
import fsleyes.actions                    as actions
import fsleyes.gl.annotations             as annotations
import fsleyes.profiles.orthoviewprofile  as orthoviewprofile


log = logging.getLogger(__name__)


class OrthoCropProfile(orthoviewprofile.OrthoViewProfile):
    """The ``OrthoCropProfile`` class is a :class:`.Profile` for the
    :class:`.OrthoPanel` class, which allows the user to define a cropping
    region for :class:`.Image` overlays.


    Ther ``OrthoCropProfile`` displays a cropping, or ROI, region on the
    ``OrthoPanel`` canvases, relative to the for the currently selected
    image, using :class:`.Rect` annotations. Mouse handlers are also
    defined, allowing the user to adjust the box.


    Once the user has selected a cropping region, the related
    :class:`.CropImagePanel` allows him/her to create a cropped copy of the
    image.


    The ``OrthoCropProfile`` class defines one mode, in addition to those
    inherited from the :class:`.OrthoViewProfile` class:


    ======== ===================================================
    ``crop`` Clicking and dragging allows the user to change the
             boundaries of a cropping region.
    ======== ===================================================


    .. note:: The crop overlay will only be shown if the
              :attr:`.DisplayContext.displaySpace` is set to the currently
              selected overlay. The :class:`.CropImagePanel` uses a
              :class:`.DisplaySpaceWarning` to inform the user.
    """


    cropBox = props.Bounds(ndims=3, real=False, minDistance=1, clamped=False)
    """This property keeps track of the current low/high limits
    of the cropping region, in the voxel coordinate system of the
    currently selected overlay.
    """


    @staticmethod
    def tempModes():
        """Returns the temporary mode map for the ``OrthoCropProfile``,
        which controls the use of modifier keys to temporarily enter other
        interaction modes.
        """
        return {
            ('crop',  wx.WXK_SHIFT)                  : 'nav',
            ('crop',  wx.WXK_CONTROL)                : 'zoom',
            ('crop',  wx.WXK_ALT)                    : 'pan',
            ('crop', (wx.WXK_CONTROL, wx.WXK_SHIFT)) : 'slice'}


    @staticmethod
    def altHandlers():
        """Returns the alternate handlers map, which allows event handlers
        defined in one mode to be re-used whilst in another mode.
        """
        return {('crop', 'MiddleMouseDrag') : ('pan', 'LeftMouseDrag')}


    def __init__(self, viewPanel, overlayList, displayCtx):
        """Create an ``OrthoCropProfile``.

        :arg viewPanel:    An :class:`.OrthoPanel` instance.
        :arg overlayList:  The :class:`.OverlayList` instance.
        :arg displayCtx:   The :class:`.DisplayContext` instance.
        """

        orthoviewprofile.OrthoViewProfile.__init__(
            self,
            viewPanel,
            overlayList,
            displayCtx,
            ['crop'])
        self.mode = 'crop'

        # The currently selected overlay,
        # and the one for which the cropping
        # box is being shown/modified.
        self.__overlay = None

        # A cache of { overlay : cropBox }
        # which stores the last cropping
        # box for a given overlay. This
        # is used to cache boxes if the
        # user selects a different overlay
        # while the crop profile is active.
        self.__cachedCrops = {}

        # axis:   one of 0, 1, or 2 (X, Y, or Z) -
        #         the voxel axis of the crop box
        #         that is being adjusted
        #
        # limits: one of 0 or 1 (lo or hi) - the
        #         low/high limit of the crop box
        #         that is being adjusted
        #
        # These fields are set when the
        # user is dragging a crop box
        # boundary
        self.__dragAxis  = None
        self.__dragLimit = None

        self.__xcanvas = viewPanel.getXCanvas()
        self.__ycanvas = viewPanel.getYCanvas()
        self.__zcanvas = viewPanel.getZCanvas()

        # A rectangle is displayed on
        # each of the canvases, showing
        # the current cropping box.
        self.__xrect   = annotations.Rect(self.__xcanvas.getAnnotations(),
                                          0, 0, 0, 0,
                                          colour=(0.3, 0.3, 1.0),
                                          alpha=30)
        self.__yrect   = annotations.Rect(self.__ycanvas.getAnnotations(),
                                          0, 0, 0, 0,
                                          colour=(0.3, 0.3, 1.0),
                                          alpha=30)
        self.__zrect   = annotations.Rect(self.__zcanvas.getAnnotations(),
                                          0, 0, 0, 0,
                                          colour=(0.3, 0.3, 1.0),
                                          alpha=30)

        displayCtx .addListener('displaySpace',
                                self.name,
                                self.__displaySpaceChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        self       .addListener('cropBox',
                                self.name,
                                self.__cropBoxChanged)

        self.__xcanvas.getAnnotations().obj(self.__xrect, hold=True)
        self.__ycanvas.getAnnotations().obj(self.__yrect, hold=True)
        self.__zcanvas.getAnnotations().obj(self.__zrect, hold=True)

        self.robustfov.enabled = fslplatform.fsldir is not None

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``OrthoCropProfile`` is no longer
        needed. Removes property listeners and does some other clean up.
        """

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.displayCtx .removeListener('displaySpace',    self.name)
        self            .removeListener('cropBox',         self.name)

        xannot = self.__xcanvas.getAnnotations()
        yannot = self.__ycanvas.getAnnotations()
        zannot = self.__zcanvas.getAnnotations()

        if xannot is not None: xannot.dequeue(self.__xrect, hold=True)
        if yannot is not None: yannot.dequeue(self.__yrect, hold=True)
        if zannot is not None: zannot.dequeue(self.__zrect, hold=True)

        orthoviewprofile.OrthoViewProfile.destroy(self)


    @actions.action
    def robustfov(self):
        """Call ``robustfov`` for the current overlay and set the
        :attr:`cropBox` based on the result.
        """

        overlay = self.__overlay

        if overlay is None:
            return

        try:
            with tempdir.tempdir():

                # in-memory image - create and
                # save a copy - use a copy
                # otherwise the original will
                # think that is has been saved
                # to disk.
                if overlay.dataSource is None:
                    overlay = fslimage.Image(overlay)
                    overlay.save('image')

                result = fslrun.runfsl('robustfov', '-i', overlay.dataSource)

                # robustfov returns two lines, the last
                # of which contains the limits, as:
                #
                #    xmin xlen ymin ylen zmin zlen
                limits = list(result.strip().split('\n')[-1].split())
                limits = [float(l) for l in limits]

                # Convert the lens to maxes
                limits[1]      += limits[0]
                limits[3]      += limits[2]
                limits[5]      += limits[4]
                self.cropBox[:] = limits

        except Exception as e:
            log.warning('Call to robustfov failed: {}'.format(str(e)))


    def __deregisterOverlay(self):
        """Called by :meth:`__selectedOverlayChanged`. Clears references
        associated with the previously selected overlay, if necessary.
        """

        if self.__overlay is None:
            return

        self.__cachedCrops[self.__overlay] = list(self.cropBox)
        self.__overlay = None


    def __registerOverlay(self, overlay):
        """Called by :meth:`__selectedOverlayChanged`. Sets up
        references associated with the given (newly selected) overlay.
        """
        self.__overlay = overlay


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` changes.
        If the overlay is a :class:`.Image` instance, it is set as the
        :attr:`.DisplayContext.displaySpace` reference, and the
        :attr:`cropBox` is configured to be relative to the newly selected
        overlay.
        """
        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is self.__overlay:
            return

        self.__deregisterOverlay()

        enabled = isinstance(overlay, fslimage.Image)

        self.__xrect.enabled = enabled
        self.__yrect.enabled = enabled
        self.__zrect.enabled = enabled

        if not enabled:
            return

        self.__registerOverlay(overlay)

        shape = overlay.shape[:3]
        crop  = self.__cachedCrops.get(overlay, None)

        if crop is None:
            crop = [0, shape[0], 0, shape[1], 0, shape[2]]

        with props.suppress(self, 'cropBox', notify=True):
            self.cropBox.xmin = 0
            self.cropBox.ymin = 0
            self.cropBox.zmin = 0
            self.cropBox.xmax = shape[0]
            self.cropBox.ymax = shape[1]
            self.cropBox.zmax = shape[2]
            self.cropBox      = crop


    def __displaySpaceChanged(self, *a):
        """Called when the :attr:`.DisplayContext.displaySpace` changes.
        Resets the :attr:`cropBox`.
        """
        cropBox     = self.cropBox
        cropBox.xlo = self.cropBox.xmin
        cropBox.ylo = self.cropBox.ymin
        cropBox.zlo = self.cropBox.zmin
        cropBox.xhi = self.cropBox.xmax
        cropBox.yhi = self.cropBox.ymax
        cropBox.zhi = self.cropBox.zmax


    def __cropBoxChanged(self, *a):
        """Called when the :attr:`cropBox` changes. Updates the :class:`.Rect`
        annotations on the :class:`.OrthoPanel` canvases.
        """

        xlo, xhi = self.cropBox.x
        ylo, yhi = self.cropBox.y
        zlo, zhi = self.cropBox.z

        xlo     -= 0.5
        ylo     -= 0.5
        zlo     -= 0.5
        xhi     -= 0.5
        yhi     -= 0.5
        zhi     -= 0.5
        coords   = np.array([
            [xlo, ylo, zlo],
            [xlo, ylo, zhi],
            [xlo, yhi, zlo],
            [xlo, yhi, zhi],
            [xhi, ylo, zlo],
            [xhi, ylo, zhi],
            [xhi, yhi, zlo],
            [xhi, yhi, zhi]])

        opts   = self.displayCtx.getOpts(self.__overlay)
        coords = opts.transformCoords(coords, 'voxel', 'display')

        mins = coords.min(axis=0)
        maxs = coords.max(axis=0)
        pads = (maxs - mins) * 0.01

        self.__xrect.x    = mins[1]
        self.__xrect.y    = mins[2]
        self.__xrect.w    = maxs[1] - mins[1]
        self.__xrect.h    = maxs[2] - mins[2]
        self.__xrect.zmin = mins[0] - pads[0]
        self.__xrect.zmax = maxs[0] + pads[0]

        self.__yrect.x    = mins[0]
        self.__yrect.y    = mins[2]
        self.__yrect.w    = maxs[0] - mins[0]
        self.__yrect.h    = maxs[2] - mins[2]
        self.__yrect.zmin = mins[1] - pads[1]
        self.__yrect.zmax = maxs[1] + pads[1]

        self.__zrect.x    = mins[0]
        self.__zrect.y    = mins[1]
        self.__zrect.w    = maxs[0] - mins[0]
        self.__zrect.h    = maxs[1] - mins[1]
        self.__zrect.zmin = mins[2] - pads[2]
        self.__zrect.zmax = maxs[2] + pads[2]

        # TODO Don't do this if you don't need to
        self.__xcanvas.Refresh()
        self.__ycanvas.Refresh()
        self.__zcanvas.Refresh()


    def __getVoxel(self, overlay, canvasPos):
        """Called by the mouse down/drag handlers. Figures out the voxel
        in the currently selected overlay which corresponds to the
        given canvas position.
        """
        vox = self.displayCtx.getOpts(overlay).getVoxel(
            canvasPos, clip=False, vround=False)
        return np.ceil(vox)


    def _cropModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Called on mouse down events. Calculates the nearest crop box
        boundary to the mouse click, adjusts the boundary accordingly,
        and saves the boundary/axis information for subsequent drag
        events (see :meth:`_cropModeLeftMouseDrag`).
        """

        copts   = canvas.opts
        overlay = self.__overlay

        if overlay is None:
            return

        # What canvas was the click on?
        if   copts.zax == 0: hax, vax = 1, 2
        elif copts.zax == 1: hax, vax = 0, 2
        elif copts.zax == 2: hax, vax = 0, 1

        # Figure out the distances from
        # the mouse click  to each crop
        # box boundary on the clicked
        # canvas
        vox        = self.__getVoxel(overlay, canvasPos)
        hlo, hhi   = self.cropBox.getLo(hax), self.cropBox.getHi(hax)
        vlo, vhi   = self.cropBox.getLo(vax), self.cropBox.getHi(vax)

        # We compare the click voxel
        # coords with each of the x/y
        # lo/hi crop box boundaries
        boundaries = np.array([
            [hlo,      vox[vax]],
            [hhi,      vox[vax]],
            [vox[hax], vlo],
            [vox[hax], vhi]])

        # As the display space is (should be) set
        # to this overlay, the display coordinate
        # system is equivalent to the scaled
        # voxel coordinate system of the
        # overlay. So we can just multiply the
        # 2D voxel coordinates by the
        # corresponding pixdims to get the
        # distances in the display coordinate
        # system.
        pixdim           = overlay.pixdim[:3]
        scVox            = [vox[hax] * pixdim[hax], vox[vax] * pixdim[vax]]
        boundaries[:, 0] = boundaries[:, 0] * pixdim[hax]
        boundaries[:, 1] = boundaries[:, 1] * pixdim[vax]

        # Calculate distance from click to
        # crop boundaries, and figure out
        # the screen axis (x/y) and limit
        # (lo/hi) to be dragged.
        dists       = (boundaries - scVox) ** 2
        dists       = np.sqrt(np.sum(dists, axis=1))
        axis, limit = np.unravel_index(np.argmin(dists), (2, 2))
        voxAxis     = [hax, vax][axis]

        axis  = int(axis)
        limit = int(limit)

        # Save these for the mouse drag handler
        self.__dragAxis  = voxAxis
        self.__dragLimit = limit

        # Update the crop box and location
        with props.suppress(self, 'cropBox', notify=True):
            self.cropBox.setLimit(voxAxis, limit, vox[voxAxis])

        self.displayCtx.location = canvasPos


    def _cropModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse drags. Updates the :attr:`cropBox` boudary
        which was clicked on (see :meth:`_cropModeLeftMouseDown`), so it
        follows the mouse location.
        """

        if self.__overlay is None or self.__dragAxis is None:
            return

        box      = self.cropBox
        axis     = self.__dragAxis
        limit    = self.__dragLimit
        oppLimit = 1 - limit
        vox      = self.__getVoxel(self.__overlay, canvasPos)

        newval = vox[axis]
        oppval = box.getLimit(axis, oppLimit)

        if   limit == 0 and newval >= oppval: newval = oppval - 1
        elif limit == 1 and newval <= oppval: newval = oppval + 1

        with props.suppress(self, 'cropBox', notify=True):
            self.cropBox.setLimit(axis, limit, newval)

        self.displayCtx.location = canvasPos


    def _cropModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse up events. Clears references used by the
        mouse down/drag handlers.
        """

        if self.__overlay is None or self.__dragAxis is None:
            return

        self.__dragAxis  = None
        self.__dragLimit = None



#!/usr/bin/env python
#
# removealloverlays.py - The RemoveAllOverlaysAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`RemoveAllOverlaysAction`, which allows the
uesr to remove all overlays from the :class:`.OverlayList`.
"""


import fsleyes.strings as strings
from . import             base
from . import             saveoverlay


class RemoveAllOverlaysAction(base.Action):
    """The ``RemoveAllOverlaysAction`` allows the uesr to remove all
    overlays from the :class:`.OverlayList`.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``RemoveAllOverlaysAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__removeAllOverlays)

        self.__name = '{}_{}'.format(type(self).__name__, id(self))

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)


    def destroy(self):
        """Must be called when this ``RemoveAllOverlaysAction`` is no longer
        needed. Removes property listeners, and then calls
        :meth:`.Action.destroy`.
        """
        self.overlayList.removeListener('overlays', self.__name)
        base.Action.destroy(self)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Updates the
        :attr:`.Action.enabled` flag
        """
        self.enabled = len(self.overlayList) > 0


    def __removeAllOverlays(self):
        """Removes all overlays from the :class:`.OverlayList`.
        """

        import wx

        allSaved = saveoverlay.checkOverlaySaveState(
            self.overlayList, self.displayCtx)

        # If there are unsaved images,
        # get the user to confirm
        if not allSaved:

            msg    = strings.messages[self, 'unsavedOverlays']
            title  = strings.titles[  self, 'unsavedOverlays']
            parent = wx.GetApp().GetTopWindow()

            dlg = wx.MessageDialog(parent,
                                   message=msg,
                                   caption=title,
                                   style=(wx.YES_NO        |
                                          wx.NO_DEFAULT    |
                                          wx.CENTRE        |
                                          wx.ICON_WARNING))

            dlg.CentreOnParent()
            if dlg.ShowModal() == wx.ID_NO:
                return

        del self.overlayList[:]
